#!/usr/bin/env python
from __future__ import with_statement
import os, sys, re
from .argparse_compat import argparse
from .custom_arguments import add_libname_arguments, update_env_with_libnames, add_logging_arguments, process_logging_arguments, LOG_ALWAYS
from .memoize import memoize
from .file_util import read_from_file, write_to_file

__all__ = ['main']

# TODO:
# - handle fake ambiguities from [Definition foo] in a comment
# - handle theorems inside proof blocks
# - do the right thing for [Theorem foo. Theorem bar. Proof. Qed. Qed.] (we do the wrong thing right now)
# - make use of glob file?

parser = argparse.ArgumentParser(description='Implement the suggestions of [Set Suggest Proof Using.]')
parser.add_argument('source', metavar='SOURCE_FILE', type=argparse.FileType('r'), nargs='?', default=sys.stdin,
                    help='the source of set suggest proof using messages; use - for stdin.')
parser.add_argument('--hide', dest='hide_reg', nargs='*', type=str,
                    default=['.*_subproof[0-9]*$'], #, '.*_Proper$'],
                    help=('Regular expressions to not display warnings about on low verbosity.  ' +
                          '[Set Suggest Proof Using] can give suggestions about hard-to-find ' +
                          'identifiers, and we might want to surpress them.'))
parser.add_argument('--no-hide', dest='hide_reg', action='store_const', const=[])
add_libname_arguments(parser)
add_logging_arguments(parser)

def mwrite_to_file(file_name, contents, do_backup=False):
    return write_to_file(file_name, contents, do_backup=do_backup, memoize=True)

REG_PROOF_USING = re.compile(r'The proof of ([^\s]+)\s*should start with(?: one of the following commands)?:((?:\s*Proof using[^\.]+\.)+)', re.MULTILINE)
REG_SUB_PROOF_USING = re.compile(r'Proof using[^\.]+\.', re.MULTILINE)

#def all_matches(reg, source, start='', **env):
#    source_text = ''
#    for i in source:
#        if source_text[:len(start)] != start:
#            source_text = ''
#        source_text += i
#        cur_match = reg.search(source_text)
#        while cur_match:
#            ignoring = source_text[:cur_match.start()].strip()
#            if ignoring:
#                env['log']('Ignoring: ' + repr(ignoring), level=3)
#            yield cur_match.groups()
#            source_text = source_text[cur_match.end():]
#            cur_match = reg.search(source_text)

def pick_suggestion(suggestions):
    for i in ('Proof using Type.', 'Proof using.', 'Proof using .', 'Proof using Type*.'):
        if i in suggestions: return i
    return suggestions[0]

def lib_to_dir_map(libnames):
    return dict((lib, dirname) for dirname, lib in libnames)

def split_to_file_and_rest(theorem_id, **kwargs):
    module_part, rest_part = theorem_id.split('#', 1)
    module_parts = module_part.split('.')
    ret = []
    rest_parts = []
    while len(module_parts) > 0:
        rest_parts.insert(0, module_parts.pop())
        if '.'.join(module_parts) in kwargs['lib_to_dir'].keys():
            dirname = kwargs['lib_to_dir']['.'.join(module_parts)]
        elif '.'.join(module_parts) == '':
            dirname = '.'
        else:
            continue
        for split_i in range(0, len(rest_parts)):
            filename = os.path.join(dirname, *(rest_parts[:split_i] + [rest_parts[split_i] + '.v']))
            if os.path.exists(filename):
                ret.append((filename, ('.'.join(rest_parts[split_i+1:]) + '#' + rest_part).strip('#')))
    return tuple(ret)

ALL_DEFINITIONS = ('Theorem', 'Lemma', 'Fact', 'Remark', 'Corollary', 'Proposition', 'Property',
                   'Definition', 'Example', 'SubClass',
                   'Let', 'Fixpoint', 'CoFixpoint',
                   'Structure', 'Coercion', 'Instance', 'Existing Instance')
ALL_DEFINITIONS_STR = (r'[ \t]*(?:' +
                      '|'.join(ALL_DEFINITIONS) +
                      r')\s+%s(?=[\s\(:{\.]|$)')

ALL_DEFINITIONS_LESS_PROPER_STR = (r'[ \t]*(?:Global\s+|Local\s+)?(?:' +
                                   r'Add\s+(?:Parametric\s+)?Morphism' +
                                   r')\s+(?:[^\.]+|\.[A-Za-z\(\)])+\.(?:\n|$)')

ALL_DEFINITIONS_QUICK = (r'Theorem', 'Lemma', 'Fact', 'Remark', 'Corollary', 'Proposition', 'Property',
                         'Definition', 'Example', 'SubClass',
                         'Let', 'Fixpoint', 'CoFixpoint'
                         'Structure', 'Coercion', 'Instance',
                         'Add Parametric Morphism')
ALL_DEFINITIONS_STR_QUICK = (r'(?:' +
                            '|'.join(ALL_DEFINITIONS_QUICK) +
                            r')\s+%s')


ALL_DEFINITIONS_FULL_STRS = (r'^([ \t]*)(' + ALL_DEFINITIONS_STR_QUICK + r'[^\.]+\.\n)',
                             r'^([ \t]*)(' + ALL_DEFINITIONS_STR_QUICK + r'(?:[^\.]+|\.[A-Za-z\(\)])+\.\n)')

ALL_DEFINITIONS_FULL_STRS_LESS_PROPER = (r'^([ \t]*)((?:Global\s+|Local\s+)?Add\s+(?:Parametric\s+)?Morphism\s+(?:[^\.]+|\.[A-Za-z\(\)])+?\s+as\s+%s\s*\.\n)',)

ALL_ENDINGS = (r'(?:Qed|Defined|Save|Admitted|Abort)\s*\.')

FOUND_BUT_UNCHANGED = object()

def get_indent(term):
    return re.findall('^\s*', term)[0]

COMMENT='comment'
STRING='string'
def get_end_of_first_line_location(term, **env):
    stack = []
    i = 0
    while i < len(term):
        env['log']('DEBUG: stack: %s, pre: %s' % (str(stack), repr(term[:i])), level=3)
        if len(stack) == 0:
            if term[i] == '.' and (len(term) == i + 1 or (term[i+1] in ' \t\r\n')):
                return i+1
            if term[i] == '.':
                while i < len(term) and term[i] == '.':
                    i += 1
            else:
                i += 1
        elif stack[-1] == STRING:
            if term[i] == '"':
                stack.pop()
            i += 1
        elif stack[-1] == COMMENT:
            if term[i:i+2] == '(*':
                stack.extend(COMMENT)
                i += 2
            elif term[i:i+2] == '*)':
                stack.pop()
                i += 2
            elif term[i] == '"':
                stack.extend(STRING)
                i += 1
            else:
                i += 1
        else: # len(stack) == 0
            assert(False)
    return len(term) - 1

def update_proof(name, before_match, match, after_match, filename, rest_id, suggestion, **env):
    ending = re.search(ALL_ENDINGS, after_match, re.MULTILINE)
    if ending:
        proof_part = after_match[:ending.start()]
        env['log']('Inspecting proof: %s' % proof_part, level=2)
        if proof_part.count('Proof') == 1:
            proof_match = re.search('Proof(?: using[^\.]*)?\.', proof_part)
            if proof_match:
                if proof_match.group() == suggestion:
                    return FOUND_BUT_UNCHANGED # already correct
                elif proof_match.group() == 'Proof.':
                    return (before_match + match.group() + after_match[:proof_match.start()] +
                            suggestion +
                            after_match[proof_match.end():])
                else:
                    env['log']('Warning: Mismatch between existing Proof using and suggested Proof using:', level=0)
                    env['log']('In %s, id %s, found %s, expected %s' % (filename, rest_id, proof_match.group(), suggestion), level=0)
                    return FOUND_BUT_UNCHANGED
            else:
                env['log']('Warning: Mismatched Proofs found in %s for %s' % (filename, rest_id), level=0)
                return FOUND_BUT_UNCHANGED
        elif proof_part.count('Proof') == 0:
            loc = get_end_of_first_line_location(proof_part, **env)
            indent = get_indent(match.group())
            env['log']('Before proof: %s\nAfter proof: %s' % (proof_part[:loc], proof_part[loc:]), level=3)
            return '%s%s%s\n%s%s\n%s%s' % (before_match, match.group(), proof_part[:loc], indent, suggestion, proof_part[loc:], after_match[ending.start():])
        else:
            env['log']('Warning: Too many Proofs found in %s for %s' % (filename, rest_id), level=0)
    else:
        env['log']('Warning: No %s found in %s for %s' % (ALL_ENDINGS, filename, rest_id), level=0)
    return None

def unsafe_update_definitions(name, contents, filename, rest_id, suggestion, **env):
    match = re.search(ALL_DEFINITIONS_STR % name, contents, re.MULTILINE)
    if match:
        return update_proof(name, contents[:match.start()], match, contents[match.end():], filename, rest_id, suggestion, **env)
    elif name[-len('_Proper'):] == '_Proper':
        for match in re.finditer(ALL_DEFINITIONS_LESS_PROPER_STR, contents, re.MULTILINE | re.DOTALL):
            if match.group().strip('\n\t .').split(' ')[-1] + '_Proper' == name:
                return update_proof(name, contents[:match.start()], match, contents[match.end():], filename, rest_id, suggestion, **env)
    env['log']('Warning: No %s found in %s' % (rest_id, filename),
               level=(2 if re.search('|'.join(env['hide_reg']), rest_id) else 0))
    return None

def update_definitions(contents, filename, rest_id, suggestion, **env):
    name = rest_id.split('#')[-1]
    if len(re.findall(ALL_DEFINITIONS_STR % name, contents, re.MULTILINE)) <= 1:
        ret = unsafe_update_definitions(name, contents, filename, rest_id, suggestion, **env)
        if ret is None:
            return None
        elif ret is FOUND_BUT_UNCHANGED:
            return contents
        else:
            return ret
    else:
        modules = rest_id.split('#')[0].split('.')
        pre, mod_body, post = '', contents, ''
        while len(modules) > 0:
            mod_name = modules.pop(0)
            cur_mod = 'Module ' + mod_name
            cur_end = 'End ' + mod_name + '.'
            if mod_body.count(cur_mod) == 1:
                pre += mod_body[:mod_body.index(cur_mod) + len(cur_mod)]
                mod_body = mod_body[mod_body.index(cur_mod) + len(cur_mod):]
                if mod_body.count(cur_end) == 1:
                    post = mod_body[mod_body.index(cur_end):] + post
                    mod_body = mod_body[:mod_body.index(cur_end)]
                    if len(re.findall(ALL_DEFINITIONS_STR % name, mod_body, re.MULTILINE)) <= 1:
                        ret = unsafe_update_definitions(name, mod_body, filename, rest_id, suggestion, **env)
                        if ret is None:
                            return None
                        elif ret is FOUND_BUT_UNCHANGED:
                            return contents
                        else:
                            return pre + ret + post
                else:
                    env['log']('Warning: Too many %s found for %s in %s' % (cur_end, rest_id, filename), level=0)
                    break
            else:
                env['log']('Warning: Too many %s found for %s in %s' % (cur_mod, rest_id, filename), level=0)
                break
        env['log']('Warning: Module disambiguation was insufficient to uniqueize %s in %s' % (rest_id, filename), level=0)
        env['log']('Found: %s' % repr(re.findall(ALL_DEFINITIONS_STR % name, contents, re.MULTILINE)), level=2)
    return contents

def main():
    args = process_logging_arguments(parser.parse_args())
    source = args.source
    env = {
        'lib_to_dir': lib_to_dir_map(args.libnames + args.non_recursive_libnames),
        'log': args.log,
        'hide_reg': args.hide_reg
        }
    update_env_with_libnames(env, args)
    for theorem_id, suggestions in REG_PROOF_USING.findall('\n'.join(source)):
        all_suggestions = REG_SUB_PROOF_USING.findall(suggestions)
        suggestion = pick_suggestion(all_suggestions)
        filenames = list(reversed(split_to_file_and_rest(theorem_id, **env)))
        if filenames:
            is_first = True
            for filename, rest_id in filenames:
                orig = read_from_file(filename)
                updated = update_definitions(orig, filename, rest_id, suggestion, **env)
                if updated is not None:
                    if updated != orig:
                        env['log']('Updating %s in %s' % (rest_id, filename), level=1)
                        mwrite_to_file(filename, updated, do_backup=True)
                    elif len(filenames) > 1 and not is_first:
                        env['log']('Found %s in %s' % (rest_id, filename), level=1)
                    break
                is_first = False
        else:
            env['log']('Warning: Could not find theorem %s' % theorem_id, level=LOG_ALWAYS)

if __name__ == '__main__':
    main()



#!/usr/bin/env python2
import shutil, os, os.path, sys
from .argparse_compat import argparse
from .import_util import get_file, sort_files_by_dependency, IMPORT_ABSOLUTIZE_TUPLE, ALL_ABSOLUTIZE_TUPLE
from .custom_arguments import add_libname_arguments, update_env_with_libnames, add_logging_arguments, process_logging_arguments
from .file_util import write_to_file

__all__ = ['main']

parser = argparse.ArgumentParser(description='Absolutize the imports of Coq files')
parser.add_argument('input_files', metavar='INFILE', nargs='*', type=argparse.FileType('r'),
                    help='.v files to update')
parser.add_argument('--in-place', '-i', metavar='SUFFIX', dest='suffix', nargs='?', type=str, default='',
                    help='update files in place (makes backup if SUFFIX supplied)')
parser.add_argument('--update-all', dest='update_all',
                    action='store_const', default=False, const=True,
                    help=("also update all .v files listed in any _CoqProject file passed to -f (implies --in-place, requires -f)"))
parser.add_argument('--all', '-a', dest='absolutize', action='store_const',
                    const=ALL_ABSOLUTIZE_TUPLE, default=IMPORT_ABSOLUTIZE_TUPLE,
                    help='Absolutize all constants, and not just imports.')
parser.add_argument('--coqbin', metavar='COQBIN', dest='coqbin', type=str, default='',
                    help='The path to a folder containing the coqc and coqtop programs.')
parser.add_argument('--coqc', metavar='COQC', dest='coqc', type=str, default='coqc',
                    help='The path to the coqc program.')
parser.add_argument('--coq_makefile', metavar='COQ_MAKEFILE', dest='coq_makefile', type=str, default='coq_makefile',
                    help='The path to the coq_makefile program.')
add_libname_arguments(parser)
add_logging_arguments(parser)


def absolutize_imports(filename, **kwargs):
    kwargs['log']('Processing %s...' % filename)
    absolutized_contents = get_file(filename, update_globs=True, **kwargs)
    if kwargs['inplace']:
        do_backup = kwargs['suffix'] is not None and len(kwargs['suffix']) > 0
        write_to_file(filename, absolutized_contents, do_backup=do_backup, backup_ext=kwargs['suffix'])
    else:
        print(absolutized_contents)

def main():
    args = process_logging_arguments(parser.parse_args())
    def prepend_coqbin(prog):
        if args.coqbin != '':
            return os.path.join(args.coqbin, prog)
        else:
            return prog
    env = {
        'log': args.log,
        'coqc': prepend_coqbin(args.coqc),
        'inplace': args.suffix != '', # it's None if they passed no argument, and '' if they didn't pass -i
        'suffix': args.suffix,
        'absolutize': args.absolutize,
        'coq_makefile': prepend_coqbin(args.coq_makefile),
        'input_files': tuple(f.name for f in args.input_files),
        }
    update_env_with_libnames(env, args)

    for f in args.input_files: f.close()

    if args.update_all:
        if env['_CoqProject'] is None:
            parser.error('--update-all given without -f')
            sys.exit(1)
        else:
            env['inplace'] = True
            if env['suffix'] == '': env['suffix'] = None
            env['input_files'] = tuple(list(env['input_files']) + list(env['_CoqProject_v_files']))
        if len(env['input_files']) == 0:
            parser.error('no .v files listed in %s' % args.CoqProjectFile)
    elif len(env['input_files']) == 0:
        parser.error('not enough arguments (-f COQPROJECTFILE with .v files is required if no .v files are given)')

    env['input_files'] = sort_files_by_dependency(env['input_files'], update_globs=True, **env)

    for f in env['input_files']:
        absolutize_imports(f, **env)

if __name__ == '__main__':
    main()



from __future__ import with_statement
import re
from .diagnose_error import get_coq_output
from .coq_version import get_coq_native_compiler_ondemand_fragment
from .coq_full_grammar import COQ_GRAMMAR_TOKENS

# Like coq_version.py, except for things that use get_coq_output (or
# perhaps a bit more generally for things that need to run coqc on a
# file)

__all__ = ["get_proof_term_works_with_time", "get_ltac_support_snippet", "get_reserved_modnames"]

def get_proof_term_works_with_time(coqc_prog, **kwargs):
    contents = r"""Lemma foo : forall _ : Type, Type.
Proof (fun x => x)."""
    output, cmds, retcode, runtime = get_coq_output(coqc_prog, ('-time', '-q'), contents, timeout_val=1, verbose_base=3, **kwargs)
    return 'Error: Attempt to save an incomplete proof' not in output

LTAC_SUPPORT_SNIPPET = {}
def get_ltac_support_snippet(coqc, **kwargs):
    if coqc in LTAC_SUPPORT_SNIPPET.keys():
        return LTAC_SUPPORT_SNIPPET[coqc]
    test = r'''Inductive False : Prop := .
Axiom proof_admitted : False.
Tactic Notation "admit" := abstract case proof_admitted.
Goal False. admit. Qed.'''
    errinfo = {}
    native_ondemand_args = list(get_coq_native_compiler_ondemand_fragment(coqc, **kwargs))
    for before, after in (('Require Coq.Init.Ltac.\n', 'Import Coq.Init.Ltac.\n'),
                          ('Require Coq.Init.Notations.\n', 'Import Coq.Init.Notations.\n'),
                          ('Declare ML Module "coq-core.plugins.ltac".\n', ''),
                          ('Declare ML Module "coq-core.plugins.ltac".\n', 'Global Set Default Proof Mode "Classic".\n'),
                          ('Declare ML Module "ltac_plugin".\n', ''),
                          ('Declare ML Module "ltac_plugin".\n', 'Global Set Default Proof Mode "Classic".\n')):
        contents = '%s\n%s\n%s' % (before, after, test)
        output, cmds, retcode, runtime = get_coq_output(coqc, tuple(['-q', '-nois'] + native_ondemand_args), contents, timeout_val=None, verbose_base=3, is_coqtop=kwargs['coqc_is_coqtop'], **kwargs)
        if retcode == 0:
            LTAC_SUPPORT_SNIPPET[coqc] = (before, after)
            return (before, after)
        else:
            errinfo[contents] = {'output': output, 'cmds': cmds, 'retcode': retcode, 'runtime': runtime}
    raise Exception('No valid ltac support snipped found.  Debugging info: %s' % repr(errinfo))

def get_is_modname_valid(coqc_prog, modname, **kwargs):
    contents = "Module %s. End %s." % (modname, modname)
    if ' ' in modname: return False
    output, cmds, retcode, runtime = get_coq_output(coqc_prog, ('-q',), contents, timeout_val=1, verbose_base=3, **kwargs)
    return 'Syntax error:' not in output

def get_reserved_modnames(coqtop_prog, **kwargs):
    grammars_contents = 'Print Grammar tactic. Print Grammar constr. Print Grammar vernac.'
    grammars_output, cmds, retcode, runtime = get_coq_output(coqtop_prog, ('-q',), grammars_contents, is_coqtop=True, pass_on_stdin=True, timeout_val=1, verbose_base=3, **kwargs)
    tokens = sorted(set([i.strip('"') for i in re.findall(r'"[a-zA-Z_][^"]*"', grammars_output)] + COQ_GRAMMAR_TOKENS))
    contents = '\n'.join("Module %s. End %s." % (modname, modname) for modname in tokens)
    output, cmds, retcode, runtime = get_coq_output(coqtop_prog, ('-q',), contents, is_coqtop=True, pass_on_stdin=True, timeout_val=10, verbose_base=3, **kwargs)
    success = re.findall(r'Module ([^ ]+) is defined', output)
    return tuple(modname for modname in tokens if modname not in success)



#!/usr/bin/env python2
from __future__ import with_statement
import os, sys, shutil, re
from .argparse_compat import argparse
from .split_file import split_coq_file_contents_with_comments
from .strip_comments import strip_comments
from .custom_arguments import add_logging_arguments, process_logging_arguments, LOG_ALWAYS
from .file_util import write_to_file
from .util import PY3
if PY3: from .util import raw_input

__all__ = ['main']

parser = argparse.ArgumentParser(description='Move various statements out of proof blocks')
parser.add_argument('input_files', metavar='INFILE', nargs='+', type=str,
                    help='.v files to update')
parser.add_argument('--in-place', '-i', metavar='SUFFIX', dest='suffix', nargs='?', type=str, default='',
                    help='update files in place (makes backup if SUFFIX supplied)')
add_logging_arguments(parser)

ALL_DEFINITIONS_REG = re.compile(r'^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?(?:' +
                                r'Theorem|Lemma|Fact|Remark|Corollary|Proposition|Property' +
                                r'|Definition|Example|SubClass' +
                                r'|Let|Fixpoint|CoFixpoint' +
                                r'|Structure|Coercion|Instance' +
                                r'|Add Parametric Morphism' +
                                r')\s', re.MULTILINE)

ONELINE_DEFINITIONS_REG = re.compile(r'^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?(?:' +
                                     r'Coercion|Existing\s+Instance' +
                                     r')\s', re.MULTILINE)

ALL_ENDINGS = re.compile(r'^(?:[}{]|\s)*(?:(?:Time|Timeout)\s+)?(?:Qed|Defined|Save|Admitted|Abort)\s*\.', re.MULTILINE)

COPY_UP_REG = re.compile(r'^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?' +
                         r'(?:Opaque\s|Transparent\s|Arguments\s|Implicit\s+Arguments\s)', re.MULTILINE)
MOVE_UP_REG = re.compile(r'^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?' +
                         r'(?:Require|Import|Notation|Ltac|Tactic\s+Notation|Infix|Delimit\s+Scope|Reserved\s+Notation|Reserved\+Infix|Existing\s+Instance|Coercion|Hint|Set\s+Printing|Unset\+Printing)\s', re.MULTILINE)
SPECIAL_TACTICS = r'W_eq|PropXRel|ILAlgoTypes|NoDup|W_neq|PropXTac|SEP_FACTS|SF\.'
SAFE_REG = re.compile(r'^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?(?:' +
                      r'[a-z]|[0-9]+\s*:|\(\s*[a-z]|' + SPECIAL_TACTICS +
                      r'|{|}|-|\+|\*' +
                      r'|(?:Unfocus|Proof|Focus|Grab\s+Existentials?|Open\s+Scope|Close\s+Scope)(?:\.|\s)|\(\*' +
                      r')', re.MULTILINE)

reg = re.compile(r'Require\s.*?\.\s+', re.MULTILINE | re.DOTALL)

def get_leading_space(string):
    return re.findall(r'^(?:\s*?\n)?([ \t]*)(?!\s)', string)[0]

def remove_leading_space(string, space_count):
    return re.sub(r'(^|\n)' + (' ' * space_count), r'\1', string, re.MULTILINE)

def set_leading_space(string, space_count):
    return remove_leading_space(string, max(0, len(get_leading_space(string)) - space_count))

def strip_parens(string):
    last = string
    cur = re.sub('"[^"]+|{[^{}]+}|\([^\(\)]+\)', '', last)
    while cur != last:
        last, cur = cur, re.sub('"[^"]+|{[^{}]+}|\([^\(\)]+\)', '', cur)
    cur = re.sub(r'\slet\s[^\(\){}"]+?:=[^\(\){}"]+?\sin\s', '', cur, re.MULTILINE)
    return cur


def space_canonicalize(stmt):
    return re.sub(r'\s+', ' ', stmt.strip(' \t\n\r.'))

def canonicalize(stmt):
    return space_canonicalize(stmt).replace('Transparent ', 'Opaque ')

def cancels(statement1, statement2):
    """Returns True if statement2 cancels the effect of statement1; False otherwise"""
    return canonicalize(statement1) == canonicalize(statement2)

def commutes(statement1, statement2):
    if any(i in statement1 or i in statement2 for i in ('(*', '*)', '"')):
        return False
    statement1, statement2 = space_canonicalize(statement1), space_canonicalize(statement2)
    m1 = re.match(r'^(?:Local |Global )?(?:Transparent|Opaque) (.*)$', statement1)
    m2 = re.match(r'^(?:Local |Global )?(?:Transparent|Opaque) (.*)$', statement2)
    if not m1 or not m2:
        return False
    constants1, constants2 = set(m1.groups()[0].split(' ')), set(m2.groups()[0].split(' '))
    return constants1.isdisjoint(constants2)

def preminimize_lifted_statements(statements):
    """Remove useless deferred statements, such as transparent followed by opaque"""

    prev = set()
    for statement in statements:
        prev = set(last for last in prev if not cancels(last, statement))
        if not all(commutes(last, statement) for last in prev):
            for last in sorted(prev):
                yield last
            prev = set()
        prev.add(statement)
    for last in sorted(prev):
        yield last


def minimize_lifted_statements(statements):
    statements = list(preminimize_lifted_statements(statements))
    sans_copy = [i for i in statements if COPY_UP_REG.match(i.strip()) is None]
    sans_copy_hint = [i for i in sans_copy if re.match(r'^(?:Local\s+|Global\s+)?Hint\s', i.strip()) is None]
    if len(sans_copy_hint) == 1:
        return sans_copy
    else:
        return statements

def move_from_proof(filename, **kwargs):
    kwargs['log']('Processing %s...' % filename)
    try:
        with open(filename, 'r') as f:
            contents = f.read()
    except IOError as e:
        kwargs['log']('Failed to process %s' % filename)
        kwargs['log'](repr(e), level=2)
        return
    ret = []
    cur_statements = []
    cur_statement = []
    deferred_copied_statements = []
    deferred_statements = []
    orig_space_count = 0
    cur_diff_space_count = 0
    if ''.join(split_coq_file_contents_with_comments(contents)) != contents:
        kwargs['log']('WARNING: Could not split %s' % filename, level=LOG_ALWAYS)
        return
    for i in split_coq_file_contents_with_comments(contents):
        is_definition_full = (ALL_DEFINITIONS_REG.match(i) is not None
                              and (':=' in strip_parens(strip_comments(i))
                                   or ONELINE_DEFINITIONS_REG.match(i)))
        is_definition_start = (ALL_DEFINITIONS_REG.match(i) is not None
                               and ':=' not in strip_parens(strip_comments(i))
                               and not ONELINE_DEFINITIONS_REG.match(i))
        #print((is_definition_start, ONELINE_DEFINITIONS_REG.match(i), ALL_DEFINITIONS_REG.match(i), i))
        is_definition_end = ALL_ENDINGS.match(i) is not None
        #if 'not_reachable_iff' in i:
        #    print((ALL_DEFINITIONS_REG.match(i), strip_parens(strip_comments(i)), ONELINE_DEFINITIONS_REG.match(i)))
        if not is_definition_start and not cur_statements and not cur_statement:
            kwargs['log'](repr(i), level=3)
            ret.append(i)
        elif is_definition_start:
            kwargs['log']('Starting definition (%d): %s' % (len(deferred_statements), repr(i)), level=2)
            if not cur_statement and not deferred_statements:
                orig_space_count = len(get_leading_space(i))
            if cur_statement:
                deferred_statements.append((cur_diff_space_count, cur_statement))
            cur_diff_space_count = max(0, len(get_leading_space(i)) - orig_space_count)
            cur_statement = [remove_leading_space(i, cur_diff_space_count)]
        elif (SAFE_REG.match(i) or not i.strip()) and cur_statement:
            kwargs['log'](repr(i), level=3)
            cur_statement.append(remove_leading_space(i, cur_diff_space_count))
        elif is_definition_end and cur_statement:
            kwargs['log']('Ending definition: ' + repr(i), level=2)
            cur_statement.append(remove_leading_space(i, cur_diff_space_count))
            cur_statements.append(''.join(cur_statement))
            #print(''.join(preminimize_lifted_statements(cur_statements)))
            if deferred_statements:
                cur_diff_space_count, cur_statement = deferred_statements.pop()
                deferred_copied_statements = list(preminimize_lifted_statements(deferred_copied_statements))
                cur_statement.extend(deferred_copied_statements) # todo: fix indentation
            else:
                #print('extending')
                #print(''.join(minimize_lifted_statements(cur_statements)))
                ret.extend(list(minimize_lifted_statements(cur_statements)))
                cur_statement = []
                cur_statements = []
                deferred_copied_statements = []
        elif MOVE_UP_REG.match(i) or is_definition_full:
            kwargs['log']('Lifting: ' + repr(i), level=2)
            cur_statements.append(set_leading_space(i, orig_space_count))
        elif COPY_UP_REG.match(i) and cur_statement:
            kwargs['log']('Lift-copying: ' + repr(i), level=2)
            cur_statement.append(remove_leading_space(i, cur_diff_space_count))
            cur_statements.append(set_leading_space(i, orig_space_count))
            deferred_copied_statements.append(remove_leading_space(i, cur_diff_space_count))
        else:
            raw_input('WARNING: Unrecognized: %s' % repr(i))
        #print(cur_diff_space_count)
        #print(remove_leading_space(i, cur_diff_space_count))
    if cur_statements or deferred_statements or cur_statement:
        raw_input('WARNING: extra statements: %s' % repr((cur_statements, cur_statement, deferred_statements)))
        cur_statements.append(''.join(cur_statement))
        for i in deferred_statements:
            cur_statements.append(''.join(i))
        ret.extend(list(minimize_lifted_statements(cur_statements)))
    ret = ''.join(ret)
    if ret == contents:
        return
    if kwargs['inplace']:
        do_backup = kwargs['suffix'] is not None and len(kwargs['suffix']) > 0
        write_to_file(filename, ret, do_backup=do_backup, backup_ext=kwargs['suffix'])
    else:
        print(ret)

def main():
    args = process_logging_arguments(parser.parse_args())
    env = {
        'log': args.log,
        'inplace': args.suffix != '', # it's None if they passed no argument, and '' if they didn't pass -i
        'suffix': args.suffix,
        }

    for f in args.input_files:
        move_from_proof(f, **env)

if __name__ == '__main__':
    main()



#!/usr/bin/env python2
import shutil, os, os.path, sys, re
from .argparse_compat import argparse
from .custom_arguments import add_libname_arguments, update_env_with_libnames, add_logging_arguments, process_logging_arguments, LOG_ALWAYS, DEFAULT_VERBOSITY
from .split_file import UnsupportedCoqVersionError
from .import_util import get_file_statements_insert_references, sort_files_by_dependency, classify_require_kind, EXPORT, REQUIRE_EXPORT
from .file_util import write_to_file
from .memoize import memoize
from .minimizer_drivers import run_binary_search
from .coq_version import DEFAULT_COQTOP
from . import diagnose_error
from . import util

__all__ = ['main']

parser = argparse.ArgumentParser(description='Remove useless Requires in a file')
parser.add_argument('input_files', metavar='INFILE', nargs='*', type=argparse.FileType('r'),
                    help='.v files to update')
parser.add_argument('--in-place', '-i', metavar='SUFFIX', dest='suffix', nargs='?', type=str, default='',
                    help='update files in place (makes backup if SUFFIX supplied)')
parser.add_argument('--update-all', '--all', dest='update_all',
                    action='store_const', default=False, const=True,
                    help=("also update all .v files listed in any _CoqProject file passed to -f (implies --in-place, requires -f)"))
parser.add_argument('--absolutize', dest='absolutize',
                    action='store_const', default=False, const=True,
                    help=("Replace Requires with fully qualified versions."))
parser.add_argument('--timeout', dest='timeout', metavar='SECONDS', type=int, default=-1,
                    help=("Use a timeout; make sure Coq is " +
                          "killed after running for this many seconds. " +
                          "If 0, there is no timeout.  If negative, then " +
                          "twice the initial run of the script is used.\n\n" +
                          "Default: -1"))
parser.add_argument('--no-keep-exports', dest='keep_exports',
                    action='store_const', default=True, const=False,
                    help=("Allow the removal of Require lines that have Export in them"))
parser.add_argument('--no-timeout', dest='timeout', action='store_const', const=0,
                    help=("Do not use a timeout"))
parser.add_argument('--keep-going', '-k', dest='keep_going', action='store_const', const=True, default=False,
                    help=("Keep going when some files can't be minimized."))
parser.add_argument('--coqbin', metavar='COQBIN', dest='coqbin', type=str, default='',
                    help='The path to a folder containing the coqc and coqtop programs.')
parser.add_argument('--coqc', metavar='COQC', dest='coqc', type=str, default='coqc',
                    help='The path to the coqc program.')
add_libname_arguments(parser)
add_logging_arguments(parser)

def mark_exports(state, keep_exports):
    return tuple((text_as_bytes, ranges, bool(keep_exports and ranges and classify_require_kind(text_as_bytes, ranges) in (EXPORT, REQUIRE_EXPORT)))
                 for text_as_bytes, ranges in state)

SKIP='SKIP'
REMOVE='REMOVE'

def trailing_whitespace(text):
    return ''.join(sorted(re.search(r'(\s*)$', text, flags=re.DOTALL).groups()[0]))

def leading_whitespace(text):
    return ''.join(sorted(re.search(r'^(\s*)', text, flags=re.DOTALL).groups()[0]))

def whitespace_to_key(ws):
    return (ws.count('\n'), ws.count('\t'), ws.count(' '), len(ws), ws)

def gobble_whitespace_text(before, after):
    before_ws, after_ws = trailing_whitespace(before), leading_whitespace(after)
    assert(whitespace_to_key('') < whitespace_to_key(' ')) # sanity check
    if before_ws and after_ws:
        if whitespace_to_key(before_ws) < whitespace_to_key(after_ws):
            return (before[:-len(before_ws)], after)
        else:
            return (before, after[len(after_ws):])
    elif before_ws:
        return (before[:-len(before_ws)], after)
    elif after_ws:
        return (before, after[len(after_ws):])
    else:
        return (before, after)

def gobble_whitespace(before, after):
    assert(before is bytes(before))
    assert(after is bytes(after))
    before, after = gobble_whitespace_text(before.decode('utf-8'), after.decode('utf-8'))
    return (before.encode('utf-8'), after.encode('utf-8'))

def step_state(state, action):
    ret = []
    state = list(state)
    while len(state) > 0:
        (text_as_bytes, references, force_keep), state = state[0], state[1:]
        if references:
            (start, end, _loc, _append, _ty), new_references = references[0], tuple(references[1:])
            if action == SKIP or action is None:
                ret.append((text_as_bytes, new_references, force_keep))
            elif action == REMOVE and not force_keep:
                if new_references: # still other imports, safe to just remove
                    pre_text_as_bytes, post_text_as_bytes = gobble_whitespace(text_as_bytes[:start], text_as_bytes[end:])
                    ret.append((pre_text_as_bytes + post_text_as_bytes, new_references, force_keep))
                else: # no other imports; remove this line completely
                    if ret and state:
                        prev_text_as_bytes, post_text_as_bytes = gobble_whitespace(ret[-1][0], state[0][0])
                        ret[-1] = (prev_text_as_bytes,) + ret[-1][1:]
                        state[0] = (post_text_as_bytes,) + state[0][1:]
                    elif ret:
                        ret[-1] = (ret[-1][0].rstrip(),) + ret[-1][1:]
                    elif state:
                        state[0] = (state[0][0].lstrip(),) + state[0][1:]
            else:
                raise ValueError
            ret.extend(state)
            return tuple(ret)
        else:
            ret.append((text_as_bytes, references, force_keep))
    return None

def state_to_contents(state):
    return ''.join(reversed([v[0].decode('utf-8') for v in state]))

# the higher verbose_base, the less verbose we are, since we start at a higher level by default
def make_check_state(original_contents, verbose_base=4-DEFAULT_VERBOSITY, **kwargs):
    # original_contents is str
    expected_output, orig_cmds, orig_retcode, runtime = diagnose_error.get_coq_output(kwargs['coqc'], kwargs['coqc_args'], original_contents, kwargs['timeout'], verbose_base=2, **kwargs)
    @memoize
    def check_contents(contents):
        output, cmds, retcode, runtime = diagnose_error.get_coq_output(kwargs['coqc'], kwargs['coqc_args'], contents, kwargs['timeout'], verbose_base=2, **kwargs)
        # TODO: Should we be checking the error message and the retcode and the output, or just the retcode?
        retval = (diagnose_error.has_error(output) or output != expected_output or retcode != orig_retcode)
        if retval:
            kwargs['log']('Failed change.  Error when running "%s":\n%s' % ('" "'.join(cmds), output), level=verbose_base-1)
        else:
            kwargs['log']('Successful change', level=verbose_base)
            kwargs['log']('New contents:\n"""\n%s\n"""' % contents, level=verbose_base+1)
        return not retval

    def check_state(state):
        return check_contents(state_to_contents(state))

    return check_state

def make_save_state(filename, **kwargs):
    def save_state(state, final=False):
        contents = state_to_contents(state)
        if kwargs['inplace']:
            do_backup = kwargs['suffix'] is not None and len(kwargs['suffix']) > 0
            write_to_file(filename, contents, do_backup=do_backup, backup_ext=kwargs['suffix'])
        elif final:
            print(contents)
    return save_state

def main():
    args = process_logging_arguments(parser.parse_args())
    env = {
        'log': args.log,
        'keep_exports': args.keep_exports,
        'keep_going': args.keep_going,
        'coqc': (args.coqc if args.coqbin == '' else os.path.join(args.coqbin, args.coqc)),
        'coqc_args': (args.coq_args if args.coq_args else tuple()),
        'timeout': args.timeout,
        'inplace': args.suffix != '', # it's None if they passed no argument, and '' if they didn't pass -i
        'suffix': args.suffix,
        'input_files': tuple(f.name for f in args.input_files),
        }
    update_env_with_libnames(env, args)

    for f in args.input_files: f.close()

    if args.update_all:
        if env['_CoqProject'] is None:
            parser.error('--update-all given without -f')
            sys.exit(1)
        else:
            env['inplace'] = True
            if env['suffix'] == '': env['suffix'] = None
            env['input_files'] = tuple(list(env['input_files']) + list(env['_CoqProject_v_files']))
        if len(env['input_files']) == 0:
            parser.error('no .v files listed in %s' % args.CoqProjectFile)
    elif len(env['input_files']) == 0:
        parser.error('not enough arguments (-f COQPROJECTFILE with .v files is required if no .v files are given)')

    for dirname, libname in env['libnames']:
        env['coqc_args'] = tuple(list(env['coqc_args']) + ['-R', dirname, libname])
    for dirname, libname in env['non_recursive_libnames']:
        env['coqc_args'] = tuple(list(env['coqc_args']) + ['-Q', dirname, libname])
    for dirname in env['ocaml_dirnames']:
        env['coqc_args'] = tuple(list(env['coqc_args']) + ['-I', dirname])

    env['input_files'] = sort_files_by_dependency(env['input_files'], update_globs=True, **env)

    try:
        failed = []
        for name in env['input_files']:
            try:
                annotated_contents = get_file_statements_insert_references(name, absolutize=(('lib',) if args.absolutize else tuple()), update_globs=True, types=('lib',), appends=('<>',), **env)
                if annotated_contents is None:
                    env['log']('ERROR: Failed to get references for %s' % name, level=LOG_ALWAYS)
                    failed.append((name, 'failed to get references'))
                    if env['keep_going']:
                        continue
                    else:
                        break
                annotated_contents = mark_exports(tuple(reversed(annotated_contents)), env['keep_exports'])
                save_state = make_save_state(name, **env)
                check_state = make_check_state(state_to_contents(annotated_contents), **env)
                verbose_check_state = make_check_state(state_to_contents(annotated_contents), verbose_base=DEFAULT_VERBOSITY, **env)
                env['log']('Running coq on initial contents...')
                if not verbose_check_state(annotated_contents):
                    env['log']('ERROR: Failed to update %s' % name, level=LOG_ALWAYS)
                    failed.append((name, 'failed to update'))
                    if env['keep_going']:
                        continue
                    else:
                        break
                valid_actions = (REMOVE,)
                final_state = run_binary_search(annotated_contents, check_state, step_state, save_state, valid_actions)
                if final_state is not None:
                    if not check_state(final_state):
                        env['log']('Internal error: Inconsistent final state on %s...' % name, level=LOG_ALWAYS)
                        failed.append((name, 'inconsistent final state'))
                        if not env['keep_going']:
                            break
                    else:
                        env['log']('Saving final version of %s...' % name)
                        save_state(final_state, final=True)
            except KeyboardInterrupt:
                raise
            except SystemExit:
                raise
            except BaseException as e:
                if env['keep_going']:
                    env['log']('Failure on %s with error %s' % (name, repr(e)), level=LOG_ALWAYS)
                    failed.append((name, e))
                else:
                    raise e
        if failed:
            env['log']('The following files failed:', level=LOG_ALWAYS)
            for name, e in failed:
                env['log'](name, level=LOG_ALWAYS)
            sys.exit(1)
    except UnsupportedCoqVersionError:
        env['log']('ERROR: Your version of coqc (%s) does not support -time' % env['coqc'], level=LOG_ALWAYS)
        sys.exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
import tempfile, sys, os, re
from . import custom_arguments
from .argparse_compat import argparse
from .import_util import get_file, lib_of_filename
from .diagnose_error import get_coq_output
from .import_util import lib_of_filename, norm_libname
from .import_util import has_dir_binding, deduplicate_trailing_dir_bindings
from .memoize import memoize
from .coq_version import get_coqc_version, get_coqtop_version, get_coqc_help, get_coq_accepts_top, group_coq_args, DEFAULT_COQTOP
from .custom_arguments import add_libname_arguments, update_env_with_libnames, add_logging_arguments, process_logging_arguments, DEFAULT_LOG, LOG_ALWAYS
from .binding_util import process_maybe_list
from .file_util import clean_v_file, read_from_file, write_to_file, restore_file
from . import util
from .util import PY3
if PY3: from .util import raw_input
from . import diagnose_error

__all__ = ['main']

parser = custom_arguments.ArgumentParser(description='List all identifiers which are not closed under the global context')
parser.add_argument('--coqbin', metavar='COQBIN', dest='coqbin', type=str, default='',
                    help='The path to a folder containing the coqc and coqtop programs.')
parser.add_argument('--coqc', metavar='COQC', dest='coqc', type=str, default='coqc',
                    help='The path to the coqc program.')
parser.add_argument('--coqc-is-coqtop', dest='coqc_is_coqtop', default=False, action='store_const', const=True,
                    help="Strip the .v and pass -load-vernac-source to the coqc programs; this allows you to pass `--coqc coqtop'")
parser.add_argument('--coqtop', metavar='COQTOP', dest='coqtop', type=str, default=DEFAULT_COQTOP,
                    help=('The path to the coqtop program (default: %s).' % DEFAULT_COQTOP))
add_libname_arguments(parser)
add_logging_arguments(parser)

def qualify_identifiers_helper(identifiers, keep_unfound=False, **kwargs):
    for ident in identifiers:
        found_file = False
        if '.' in ident:
            parts = ident.split('.')
            for i in range(len(parts)-1,0,-1):
                if not found_file:
                    old_libname = '.'.join(parts[:i])
                    new_libname = norm_libname(old_libname, **kwargs)
                    if old_libname != new_libname:
                        yield (ident, new_libname + '.' + '.'.join(parts[i:]))
                        found_file = True
        if keep_unfound and not found_file:
            yield (ident, ident)

def qualify_identifiers(identifiers, keep_unfound=False, **kwargs):
    for ident, new_ident in qualify_identifiers_helper(identifiers, keep_unfound=keep_unfound, **kwargs):
        yield new_ident

def filter_local_identifiers(identifiers, keep_unfound=False, **kwargs):
    for ident, new_ident in qualify_identifiers_helper(identifiers, keep_unfound=keep_unfound, **kwargs):
        yield ident

def get_constant_name_from_locate(val):
    val = val.strip()
    for ctype in ('Constant', 'Constructor', 'Inductive', ''):
        if val and val[:len(ctype)] == ctype:
            val = val[len(ctype):].strip()
            if '\n' in val: val = val[:val.index('\n')].strip()
            return (val, ctype)

def fix_identifiers(idents, libname):
    for ident in idents:
        if ident[:len('Warning:')] == 'Warning:':
            raw_input(ident)
            continue
        elif '.' not in ident:
            yield ident
        else:
            idx = ident.index('.')
            overlap, rest = ident[:idx], ident[idx:]
            if libname[-(idx+1):] == '.' + overlap:
                yield libname + rest
            else:
                #raw_input((ident, libname))
                #yield ident
                continue

def main():
    try:
        args = process_logging_arguments(parser.parse_args())
    except argparse.ArgumentError as exc:
        if exc.message == 'expected one argument':
            exc.reraise('\nNote that argparse does not accept arguments with leading dashes.\nTry --foo=bar or --foo " -bar", if this was your intent.\nSee Python issue 9334.')
        else:
            exc.reraise()
    def prepend_coqbin(prog):
        if args.coqbin != '':
            return os.path.join(args.coqbin, prog)
        else:
            return prog
    env = {
        'log': args.log,
        'coqc': prepend_coqbin(args.coqc),
        'coqtop': prepend_coqbin(args.coqtop),
        'coqc_args': tuple(i.strip()
                           for i in process_maybe_list(args.coq_args, log=args.log)),
        'coqc_is_coqtop': args.coqc_is_coqtop,
        'temp_file_name': '',
    }

    env['remove_temp_file'] = False
    if env['temp_file_name'] == '':
        temp_file = tempfile.NamedTemporaryFile(suffix='.v', dir='.', delete=False)
        env['temp_file_name'] = temp_file.name
        temp_file.close()
        env['remove_temp_file'] = True

    if env['coqc_is_coqtop']:
        if env['coqc'] == 'coqc': env['coqc'] = env['coqtop']
        env['make_coqc'] = util.resource_path('coqtop-as-coqc.sh') + ' ' + env['coqc']

    coqc_help = get_coqc_help(env['coqc'], **env)
    coqc_version = get_coqc_version(env['coqc'], **env)

    if has_dir_binding(env['coqc_args'], coqc_help=coqc_help):
        update_env_with_libnames(env, args, default=tuple([]))
    else:
        update_env_with_libnames(env, args)

    for dirname, libname in env['libnames']:
        env['coqc_args'] = tuple(list(env['coqc_args']) + ['-R', dirname, libname])
    for dirname, libname in env['non_recursive_libnames']:
        env['coqc_args'] = tuple(list(env['coqc_args']) + ['-Q', dirname, libname])
    for dirname in env['ocaml_dirnames']:
        env['coqc_args'] = tuple(list(env['coqc_args']) + ['-I', dirname])
    env['coqc_args'] = deduplicate_trailing_dir_bindings(env['coqc_args'], coqc_help=coqc_help, coq_accepts_top=get_coq_accepts_top(env['coqc']))

    try:
        env['log']('Listing identifiers...')
        unknown = []
        closed_idents = []
        open_idents = []
        errors = []
        ignore_header = ('Welcome to Coq', '[Loading ML file')
        error_header = ('Toplevel input, characters',)
        for filename in sorted(env['_CoqProject_v_files']):
            env['log']('Qualifying identifiers in %s...' % filename, level=2)
            env['log']('Printing assumptions in %s...' % filename, max_level=2)
            libname = lib_of_filename(filename, **env)
            require_statement = 'Require %s.\n' % libname
            search_code = r"""%s
Set Search Output Name Only.
SearchPattern _ inside %s.""" % (require_statement, libname)
            output, cmds, retcode, runtime = get_coq_output(env['coqc'], env['coqc_args'], search_code, 0, is_coqtop=env['coqc_is_coqtop'], verbose_base=3, **env)
            identifiers = sorted(fix_identifiers(set(i.strip() for i in output.split('\n') if i.strip()), libname))
            print_assumptions_code = require_statement + '\n'.join('Locate %s.\nPrint Assumptions %s.' % (i, i) for i in identifiers)
            env['log']('Printing assumptions...', level=2)
            output, cmds, retcode, runtime = get_coq_output(env['coqtop'], env['coqc_args'], print_assumptions_code, 0, is_coqtop=True, pass_on_stdin=True, verbose_base=3, **env)
            i = 0
            statements = output.split('\nCoq <')
            while i < len(statements):
                if i+1 < len(statements) and ' Closed under the global context ' in statements[i+1].replace('\n', ' '):
                    last, ctype = get_constant_name_from_locate(statements[i])
                    closed_idents.append((last, ctype))
                    env['log']('Closed: %s (%s)' % (last, ctype), level=2)
                    i += 2
                elif i+1 < len(statements) and 'Axioms:' in statements[i+1].replace('\n', ' '):
                    last, ctype = get_constant_name_from_locate(statements[i])
                    open_idents.append((last, ctype, statements[i+1]))
                    env['log']('OPEN: %s (%s)' % (last, ctype))
                    env['log'](statements[i+1], level=3)
                    i += 2
                elif not statements[i].strip():
                    i += 1
                else:
                    found_ignore, found_error = False, False
                    for header in ignore_header:
                        if statements[i].strip()[:len(header)] == header:
                            found_ignore = True
                            env['log']('Ignoring: %s' % statements[i], level=3)
                    for header in ignore_header:
                        if statements[i].strip()[:len(header)] == header:
                            found_error = True
                    if not found_ignore:
                        if found_error:
                            errors.append(statements[i])
                            env['log']('ERROR: %s' % statements[i])
                        else:
                            env['log']('UNKNOWN: %s' % statements[i])
                            unknown.append(statements[i])
                    i += 1
        env['log']('Identifiers which are not closed under the global context:\n%s' % '\n'.join(ident[0] for ident in open_idents),
                   level=LOG_ALWAYS)
    finally:
        if env['remove_temp_file']:
            clean_v_file(env['temp_file_name'])

if __name__ == '__main__':
    main()



import os, sys, re

__all__ = ["prompt", "yes_no_prompt", "b", "s", "cmp_compat", "PY3", "raw_input", "re_escape", "slice_string_at_bytes", "len_in_bytes", "shlex_quote", "resource_path"]

SCRIPT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

if sys.version_info < (3,):
    PY3 = False
    def b(x):
        return x
    def s(x):
        return x
else:
    PY3 = True
    def b(x):
        if x is not None: return x.encode()
    def s(x):
        # Sometimes we get str rather than bytes??? cf https://gitlab.com/coq/coq/-/jobs/544269051
        if hasattr(x, 'decode'): return x.decode('utf-8', 'ignore')
        return x
    def cmp(x, y):
        if x < y: return -1
        if y < x: return 1
        return 0
    raw_input = input

if sys.version_info < (3, 7):
    # see https://bugs.python.org/issue29995
    def re_escape(*args, **kwargs):
        ret = re.escape(*args, **kwargs)
        for ch in ':"':
            ret = ret.replace('\\' + ch, ch)
        return ret
else:
    re_escape = re.escape

cmp_compat = cmp

def prompt(options, case_sensitive=False, strip=True, sep='/', prefix='Please enter ', postfix=': ', yes_value=True, yes=False):
    msg = prefix + sep.join(i['display'] for i in options) + postfix
    if yes:
        print(msg)
        return yes_value
    else:
        response = raw_input(msg)
        while True:
            if not case_sensitive:
                response = response.lower()
            if strip:
                response = response.strip()
            for expected in options:
                if response in expected['values']:
                    return expected['value']
            print('Response "%s" is not one of %s' % (response, ', '.join(repr(val) for i in options for val in i['values'])))
            response = raw_input(msg)

def yes_no_prompt(**kwargs):
    return prompt(({'value':True, 'display':'(y)es', 'values':('y', 'yes')},
                   {'value':False, 'display':'(n)o', 'values':('n', 'no')}),
                  yes_value=True,
                  **kwargs)

# Unfortunately, coqtop -emacs -time reports character locations in
# bytes, as does the glob file, so we need to handle unicode.  Here we
# write a generic slicer based on unicode locations that works across
# versions of python
def slice_string_at_bytes(string, start=None, end=None):
    string = b(string)
    if start is None: start = 0
    if end is None: end = len(string)
    return s(string[start:end])

def len_in_bytes(string):
    return len(b(string))

def normalize_newlines(string):
    return string.replace('\r\n', '\n').replace('\r', '\n')

# Terminal colors (maybe something cleverer needs to be done for other
# platforms).
class colors:
  ESC ='\033'
  #Escape code doesn't render on github so we use the standard escaped escape.
  #ESC = ''

  HEADER    = ESC + '[95m'
  OKBLUE    = ESC + '[94m'
  OKCYAN    = ESC + '[96m'
  OKGREEN   = ESC + '[92m'
  WARNING   = ESC + '[93m'
  FAIL      = ESC + '[91m'
  ENDC      = ESC + '[0m'
  BOLD      = ESC + '[1m'
  UNDERLINE = ESC + '[4m'

# Colors a string a given color
# Example usage: color("Hello World!", colors.OKBLUE)
def color(str, col, on=True):
    if not on:
      return str;
    else:
      return col + str + colors.ENDC;

if sys.version_info < (3, 3):
    import pipes
    shlex_quote = pipes.quote
else:
    import shlex
    shlex_quote = shlex.quote

def resource_path(path):
    if os.path.exists(os.path.join(SCRIPT_DIRECTORY, path)):
        return os.path.join(SCRIPT_DIRECTORY, path)
    if sys.version_info < (3, 7):
        import pkg_resources

        return pkg_resources.resource_filename('coq_tools', path)
    else:
        import importlib.resources

        return importlib.resources.path('coq_tools', path)



# autogenerated by cat /path/to/coq/doc/tools/docgram/fullGrammar | ./generate-full-grammar.py > coq_full_grammar.py
__all__ = ["COQ_GRAMMAR_TOKENS"]

COQ_GRAMMAR_TOKENS = ['Abort',
'About',
'Add',
'Admit',
'Admitted',
'All',
'Arguments',
'As',
'Assumptions',
'Axiom',
'Axioms',
'Back',
'BackTo',
'BinOp',
'BinOpSpec',
'BinRel',
'Bind',
'Blacklist',
'Boolean',
'Canonical',
'Case',
'Cd',
'Check',
'Class',
'Classes',
'Close',
'CoFixpoint',
'CoInductive',
'Coercion',
'Coercions',
'Collection',
'Combined',
'Comments',
'Compute',
'Conjecture',
'Conjectures',
'Constant',
'Constants',
'Constraint',
'Constructors',
'Context',
'Corollary',
'Create',
'CstOp',
'Cumulative',
'Custom',
'Cut',
'CutOff',
'Debug',
'Declare',
'Defined',
'Definition',
'Delimit',
'Dependencies',
'Dependency',
'Dependent',
'Derive',
'Diffs',
'Drop',
'Elimination',
'End',
'Entry',
'Equality',
'Equivalent',
'Eval',
'Example',
'Existentials',
'Existing',
'Export',
'Extern',
'Extra',
'Extract',
'Extraction',
'Fact',
'Fail',
'Field',
'Fields',
'File',
'Firstorder',
'Fixpoint',
'Flags',
'Focus',
'Format',
'From',
'Funclass',
'Function',
'Functional',
'GC',
'Generalizable',
'Generate',
'Global',
'Goal',
'Grammar',
'Graph',
'Guarded',
'Haskell',
'Heap',
'Hint',
'HintDb',
'Hints',
'Hypotheses',
'Hypothesis',
'Identity',
'Immediate',
'Implicit',
'Implicits',
'Import',
'Include',
'Induction',
'Inductive',
'Infix',
'Info',
'Initial',
'InjTyp',
'Inline',
'Inlined',
'Inspect',
'Instance',
'Instances',
'Intro',
'Intros',
'Inversion',
'Inversion_clear',
'JSON',
'Keys',
'Language',
'Left',
'Lemma',
'Let',
'Lia',
'Libraries',
'Library',
'Load',
'LoadPath',
'Local',
'Locate',
'Ltac',
'Ltac2',
'ML',
'Match',
'Method',
'Minimality',
'Mode',
'Module',
'Modules',
'Monomorphic',
'Morphism',
'Namespace',
'Next',
'No',
'NoInline',
'NonCumulative',
'Notation',
'Number',
'OCaml',
'Obligation',
'Obligations',
'Off',
'On',
'Opaque',
'Open',
'Optimize',
'Options',
'Parameter',
'Parameters',
'Parametric',
'Path',
'Paths',
'Polymorphic',
'Prenex',
'Preterm',
'Primitive',
'Print',
'Private',
'Profile',
'Program',
'Projections',
'Proof',
'Prop',
'PropBinOp',
'PropOp',
'PropUOp',
'Property',
'Proposition',
'Pwd',
'Qed',
'Quit',
'Rec',
'Record',
'Recursive',
'Redirect',
'Reduction',
'Register',
'Registered',
'Relation',
'Remark',
'Remove',
'Require',
'Reserved',
'Reset',
'Resolve',
'Restart',
'Rewrite',
'Right',
'Ring',
'Rings',
'SProp',
'Saturate',
'Save',
'Scheme',
'Scope',
'Scopes',
'Search',
'SearchPattern',
'SearchRewrite',
'Section',
'Separate',
'Set',
'Setoid',
'Show',
'Signatures',
'Solve',
'Solver',
'Sort',
'Sortclass',
'Sorted',
'Step',
'Strategies',
'Strategy',
'String',
'Structure',
'SubClass',
'Subgraph',
'Succeed',
'SuchThat',
'Table',
'Tables',
'Tactic',
'Term',
'Test',
'TestCompile',
'Theorem',
'Time',
'Timeout',
'To',
'Transparent',
'Type',
'TypeClasses',
'Typeclasses',
'Types',
'Typing',
'UnOp',
'UnOpSpec',
'Undelimit',
'Undo',
'Unfocus',
'Unfocused',
'Unfold',
'Universe',
'Universes',
'Unset',
'Unshelve',
'Variable',
'Variables',
'Variant',
'Verbose',
'View',
'Visibility',
'Zify',
'_',
'abstract',
'absurd',
'admit',
'after',
'all',
'and',
'any',
'apply',
'as',
'assert',
'assert_succeeds',
'associativity',
'assumption',
'at',
'auto',
'autoapply',
'autorewrite',
'autounfold',
'autounfold_one',
'before',
'best_effort',
'beta',
'bfs',
'bidirectionality',
'bigint',
'binder',
'bottom',
'bottomup',
'btauto',
'by',
'case',
'casetype',
'cbn',
'cbv',
'change',
'change_no_check',
'choice',
'clear',
'clearbody',
'closed',
'cofix',
'compare',
'completeness',
'compute',
'concl',
'congr',
'congruence',
'constants',
'constr',
'constr_eq',
'constr_eq_nounivs',
'constr_eq_strict',
'constructor',
'context',
'contradiction',
'custom',
'cut',
'cutoff',
'cutrewrite',
'cycle',
'debug',
'decidable',
'decide',
'decompose',
'default',
'delta',
'dependent',
'destauto',
'destruct',
'dfs',
'discriminate',
'discriminated',
'div',
'do',
'eapply',
'eassert',
'eassumption',
'eauto',
'ecase',
'econstructor',
'edestruct',
'ediscriminate',
'eelim',
'eenough',
'eexact',
'eexists',
'einduction',
'einjection',
'eintros',
'eleft',
'elim',
'elimtype',
'else',
'emp',
'end',
'enough',
'epose',
'eps',
'eqn',
'equality',
'eremember',
'erewrite',
'eright',
'eset',
'esimplify_eq',
'esplit',
'etransitivity',
'eval',
'evar',
'exact',
'exact_no_check',
'exactly_once',
'exists',
'expand',
'external',
'extra',
'f_equal',
'fail',
'field_lookup',
'finish_timing',
'first',
'firstorder',
'fix',
'fold',
'for',
'forall',
'format',
'fresh',
'from',
'fun',
'functional',
'gen',
'generalize',
'generalize_eqs',
'generalize_eqs_vars',
'generally',
'gfail',
'gintuition',
'give_up',
'global',
'goal',
'graph',
'guard',
'has_evar',
'have',
'head',
'head_of_constr',
'headconcl',
'headhyp',
'hget_evar',
'hint',
'hints',
'hnf',
'hresolve_core',
'hyp',
'id',
'ident',
'idtac',
'if',
'implicits',
'in',
'induction',
'infoH',
'info_auto',
'info_eauto',
'info_trivial',
'injection',
'inline',
'innermost',
'inside',
'instantiate',
'into',
'intro',
'intros',
'inversion',
'inversion_clear',
'iota',
'is',
'is_cofix',
'is_const',
'is_constructor',
'is_evar',
'is_fix',
'is_ground',
'is_ind',
'is_proj',
'is_var',
"isn't",
'lapply',
'last',
'lazy',
'lazymatch',
'left',
'let',
'level',
'loss',
'ltac',
'ltac1',
'ltac1val',
'mapping',
'match',
'max',
'measure',
'morphism',
'move',
'multimatch',
'mutable',
'name',
'native_cast_no_check',
'native_compute',
'never',
'next',
'no',
'nomatch',
'not_evar',
'notypeclasses',
'nsatz_compute',
'numgoals',
'of',
'old_hints',
'once',
'only',
'opaque',
'open_constr',
'optimize_heap',
'outermost',
'outside',
'par',
'parsing',
'pat',
'pattern',
'pose',
'postprocess',
'power',
'power_tac',
'preprocess',
'printing',
'profile',
'profiling',
'progress',
'progress_evars',
'proof',
'protect_fv',
'proved',
'rec',
'record',
'red',
'reference',
'refine',
'refl',
'reflexivity',
'remember',
'removed',
'rename',
'repeat',
'replace',
'reset',
'restart_timer',
'return',
'reverse',
'revert',
'revgoals',
'rewrite',
'rewrite_db',
'rewrite_strat',
'right',
'ring_lookup',
'rtauto',
'scope',
'scopes',
'set',
'setoid',
'setoid_etransitivity',
'setoid_reflexivity',
'setoid_rewrite',
'setoid_symmetry',
'setoid_transitivity',
'shelve',
'shelve_unifiable',
'show',
'sign',
'signature',
'simpl',
'simple',
'simplify_eq',
'soft',
'solve',
'solve_constraints',
'specialize',
'specialize_eqs',
'split',
'ssrinstancesofruleL2R',
'ssrinstancesofruleR2L',
'ssrinstancesoftpat',
'start',
'stepl',
'stepr',
'stop',
'strict',
'struct',
'subst',
'substitute',
'subterm',
'subterms',
'suff',
'suffices',
'sum',
'swap',
'symmetry',
'terms',
'then',
'time',
'timeout',
'top',
'topdown',
'transitivity',
'transparent',
'transparent_abstract',
'trivial',
'try',
'tryif',
'type',
'type_term',
'typeclasses',
'under',
'unfold',
'unify',
'unlock',
'unshelve',
'until',
'using',
'value',
'via',
'vm_cast_no_check',
'vm_compute',
'warning',
'wf',
'where',
'with',
'with_strategy',
'without',
'wlia',
'wlog',
'wlra_Q',
'wnia',
'wnra_Q',
'wpsatz_Q',
'wpsatz_Z',
'wsos_Q',
'wsos_Z',
'xlia',
'xlra_Q',
'xlra_R',
'xnia',
'xnra_Q',
'xnra_R',
'xpsatz_Q',
'xpsatz_R',
'xpsatz_Z',
'xsos_Q',
'xsos_R',
'xsos_Z',
'zeta',
'zify_elim_let',
'zify_iter_let',
'zify_iter_specs',
'zify_op',
'zify_saturate']



#!/usr/bin/env python2
from __future__ import with_statement
import os, sys, shutil, re
from .argparse_compat import argparse

__all__ = ['main']

parser = argparse.ArgumentParser(description='Move the [Require] statements up to the top of a file')
parser.add_argument('input_files', metavar='INFILE', nargs='+', type=str,
                    help='.v files to update')
parser.add_argument('--in-place', '-i', metavar='SUFFIX', dest='suffix', nargs='?', type=str, default='',
                    help='update files in place (makes backup if SUFFIX supplied)')
parser.add_argument('--verbose', '-v', dest='verbose',
                    action='count',
                    help='display some extra information')
parser.add_argument('--quiet', '-q', dest='quiet',
                    action='count',
                    help='the inverse of --verbose')
parser.add_argument('--log-file', '-l', dest='log_files', nargs='*', type=argparse.FileType('w'),
                    default=[sys.stderr],
                    help='The files to log output to.  Use - for stdout.')

DEFAULT_VERBOSITY=1

def make_logger(log_files):
    def log(text):
        for i in log_files:
            i.write(str(text) + '\n')
            i.flush()
            if i.fileno() > 2: # stderr
                os.fsync(i.fileno())
    return log

DEFAULT_LOG = make_logger([sys.stderr])

def backup(file_name, ext='.bak'):
    if not ext:
        raise ValueError
    if os.path.exists(file_name):
        backup(file_name + ext)
        os.rename(file_name, file_name + ext)

def write_to_file(file_name, contents, do_backup=False, ext='.bak'):
    if do_backup:
        backup(file_name, ext=ext)
    try:
        with open(file_name, 'w', encoding='UTF-8') as f:
            f.write(contents)
    except TypeError:
        with open(file_name, 'w') as f:
            f.write(contents)

reg = re.compile(r'Require\s.*?\.\s+', re.MULTILINE | re.DOTALL)
def move_requires(filename, **kwargs):
    if kwargs['verbose']: kwargs['log']('Processing %s...' % filename)
    with open(filename, 'r') as f:
        contents = f.read()
    requires = reg.findall(contents)
    if len(requires) == 0:
        if 'Require' not in contents:
            return
        else:
            kwargs['log']('Warning: Inconsistent [Require] in %s' % filename)
    header_end = contents.index(requires[0])
    header, rest = contents[:header_end], contents[header_end:]
    rest = reg.sub('', rest)
    new_contents = header + '\n'.join(map(str.strip, requires)) + '\n\n' + rest
    if kwargs['inplace']:
        do_backup = kwargs['suffix'] is not None and len(kwargs['suffix']) > 0
        write_to_file(filename, new_contents, do_backup=do_backup, ext=kwargs['suffix'])
    else:
        print(new_contents)

def main():
    args = parser.parse_args()
    if args.verbose is None: args.verbose = DEFAULT_VERBOSITY
    if args.quiet is None: args.quiet = 0
    verbose = args.verbose - args.quiet
    log = make_logger(args.log_files)
    env = {
        'verbose': verbose,
        'log': log,
        'inplace': args.suffix != '', # it's None if they passed no argument, and '' if they didn't pass -i
        'suffix': args.suffix,
        }

    for f in args.input_files:
        move_requires(f, **env)

if __name__ == '__main__':
    main()



import re, time
from .Popen_noblock import Popen_async, PIPE, STDOUT, Empty

__all__ = ["join_definitions", "split_statements_to_definitions"]

DEFAULT_VERBOSITY=1

def DEFAULT_LOG(text, level=DEFAULT_VERBOSITY):
    if level <= DEFAULT_VERBOSITY:
        print(text)

def get_all_nowait_iter(q):
    try:
        while True:
            yield q.get_nowait()
    except Empty:
        pass

def get_all_nowait(q):
    return ''.join(get_all_nowait_iter(q))

def get_all_semiwait_iter(q, log=DEFAULT_LOG):
    def log_and_return(val):
        log(val, level=5)
        return val
    try:
        # this is blocking; TODO(jgross): Figure out how to get coqtop
        # to tell us if it's finished computing
        yield log_and_return(q.get(True))
        while True:
            yield log_and_return(q.get(True, 0.1))
    except Empty:
        pass

def get_all_semiwait(q, log=DEFAULT_LOG):
    return ''.join(get_all_semiwait_iter(q, log=log))

def get_definitions_diff(previous_definition_string, new_definition_string):
    """Returns a triple of lists (definitions_removed,
    definitions_shared, definitions_added)"""
    old_definitions = [i for i in previous_definition_string.split('|') if i]
    new_definitions = [i for i in new_definition_string.split('|') if i]
    return (tuple(i for i in old_definitions if i not in new_definitions),
            tuple(i for i in old_definitions if i in new_definitions),
            tuple(i for i in new_definitions if i not in old_definitions))


#def split_coq_code_to_definitions(code, log=DEFAULT_LOG, coqtop='coqtop'):
#    """Splits coq code into chunks which make up
#    independent definitions/hints/etc, of the form
#
#    {
#     'statements': <list of runnable statements>,
#     'statement':  <entire chunk of code>,
#     'terms_defined': <tuple of terms defined by this chunk of code>,
#     'times': <a list of times for each statement>
#     'output': <a string of the response from coqtop>
#    }"""
#    p = Popen_async([coqtop, '-q', '-emacs', '-time'], stdout=PIPE, stderr=STDOUT, stdin=PIPE)
#    time.sleep(1)


def split_statements_to_definitions(statements, log=DEFAULT_LOG, coqtop='coqtop', coqtop_args=tuple()):
    """Splits a list of statements into chunks which make up
    independent definitions/hints/etc."""
    p = Popen_async([coqtop, '-q', '-emacs'] + list(coqtop_args), stdout=PIPE, stderr=STDOUT, stdin=PIPE)
    time.sleep(1)
    prompt_reg = re.compile(r'<prompt>([^<]*?) < ([0-9]+) ([^<]*?) ([0-9]+) < ([^<]*?)</prompt>'.replace(' ', r'\s*'))
    defined_reg = re.compile(r'^([^\s]+) is (?:defined|assumed)$', re.MULTILINE)
    # aborted_reg = re.compile(r'^Current goal aborted$', re.MULTILINE)
    # goal_reg = re.compile(r'^\s*=+\s*$', re.MULTILINE)
    # goals and definitions are on stdout, prompts are on stderr
    # clear stdout
    get_all_semiwait(p.stdout, log=log)
    # clear stderr
    # get_all_nowait(p.stderr)

    rtn = []
    cur_definition = {}
    last_definitions = '||'
    cur_definition_names = '||'
    for statement in statements:
        if not statement.strip():
            continue
        log('Write: %s\n\nWait to read...' % statement, level=4)
        p.stdin.write(statement + '\n\n')
        p.stdin.flush()
        stdout = get_all_semiwait(p.stdout, log=log)
        stderr = stdout # ''.join(get_all_semiwait(p.stderr))

        terms_defined = defined_reg.findall(prompt_reg.sub('', stdout))
        # print((statement, stdout, terms_defined))
        prompt_match = prompt_reg.search(stderr)

        if not prompt_match or len(prompt_match.groups()) != 5:
            if not prompt_match:
                log('Likely fatal warning: I did not recognize the output from coqtop:')
                log('stdout: %s\nstderr: %s' % (repr(stdout), repr(stderr)))
                log("I will append the current statement (%s) to the list of definitions as-is, but I don't expect this to work." % statement)
            else:
                log("Crazy things are happening; the number of groups isn't what it should be (should be 5 groups):")
                log("prompt_match.groups(): %s\nstdout: %s\nstderr: %s\nstatement: %s\n" % (repr(prompt_match.groups()), repr(stdout), repr(stderr), repr(statement)))

            log((statement, terms_defined, cur_definition_names, cur_definition.get(cur_definition_names, [])), level=2)
            if cur_definition_names.strip('|'):
                cur_definition[cur_definition_names]['statements'].append(statement)
                cur_definition[cur_definition_names]['terms_defined'] += terms_defined
            else:
                rtn.append({'statements':(statement,),
                            'statement':statement,
                            'terms_defined':tuple(terms_defined)})
        else:
            cur_name, line_num1, cur_definition_names, line_num2, unknown = prompt_reg.search(stderr).groups()
            definitions_removed, definitions_shared, definitions_added = get_definitions_diff(last_definitions, cur_definition_names)

            # first, to be on the safe side, we add the new
            # definitions key to the dict, if it wasn't already there.
            if cur_definition_names.strip('|') and cur_definition_names not in cur_definition:
                cur_definition[cur_definition_names] = {'statements':[], 'terms_defined':[]}


            log((statement, terms_defined, last_definitions, cur_definition_names, cur_definition.get(last_definitions, []), cur_definition.get(cur_definition_names, [])), level=2)


            # first, we handle the case where we have just finished
            # defining something.  This should correspond to
            # len(definitions_removed) > 0 and len(terms_defined) > 0.
            # If only len(definitions_removed) > 0, then we have
            # aborted something.  If only len(terms_defined) > 0, then
            # we have defined something with a one-liner.
            if definitions_removed:
                cur_definition[last_definitions]['statements'].append(statement)
                cur_definition[last_definitions]['terms_defined'] += terms_defined
                if cur_definition_names.strip('|'):
                    # we are still inside a definition.  For now, we
                    # flatten all definitions.
                    #
                    # TODO(jgross): Come up with a better story for
                    # nested definitions.
                    cur_definition[cur_definition_names]['statements'] += cur_definition[last_definitions]['statements']
                    cur_definition[cur_definition_names]['terms_defined'] += cur_definition[last_definitions]['terms_defined']
                    del cur_definition[last_definitions]
                else:
                    # we're at top-level, so add this as a new
                    # definition
                    rtn.append({'statements':tuple(cur_definition[last_definitions]['statements']),
                                'statement':'\n'.join(cur_definition[last_definitions]['statements']),
                                'terms_defined':tuple(cur_definition[last_definitions]['terms_defined'])})
                    del cur_definition[last_definitions]
                    #print('Adding:')
                    #print(rtn[-1])
            elif terms_defined:
                if cur_definition_names.strip('|'):
                    # we are still inside a definition.  For now, we
                    # flatten all definitions.
                    #
                    # TODO(jgross): Come up with a better story for
                    # nested definitions.
                    cur_definition[cur_definition_names]['statements'].append(statement)
                    cur_definition[cur_definition_names]['terms_defined'] += terms_defined
                else:
                    # we're at top level, so add this as a new
                    # definition
                    rtn.append({'statements':(statement,),
                                'statement':statement,
                                'terms_defined':tuple(terms_defined)})

            # now we handle the case where we have just opened a fresh
            # definition.  We've already added the key to the
            # dictionary.
            elif definitions_added:
                #print(definitions_added)
                cur_definition[cur_definition_names]['statements'].append(statement)
            else:
                # if we're in a definition, append the statement to
                # the queue, otherwise, just add it as it's own
                # statement
                if cur_definition_names.strip('|'):
                    cur_definition[cur_definition_names]['statements'].append(statement)
                else:
                    rtn.append({'statements':(statement,),
                                'statement':statement,
                                'terms_defined':tuple()})
        last_definitions = cur_definition_names

    log((last_definitions, cur_definition_names), level=2)
    if last_definitions.strip('||'):
        rtn.append({'statements':tuple(cur_definition[cur_definition_names]['statements']),
                    'statement':'\n'.join(cur_definition[cur_definition_names]['statements']),
                    'terms_defined':tuple(cur_definition[cur_definition_names]['terms_defined'])})
        del cur_definition[last_definitions]
    #for i in rtn:
        #print(i)
    p.stdin.close()
    return rtn

def join_definitions(definitions):
    return '\n'.join(i['statement'] for i in definitions)



import re

__all__ = ['strip_newlines']

def strip_newlines(contents, max_consecutive_newlines):
    """Removes consecutive newlines in excess of max_consecutive_newlines.

    If max_consecutive_newlines < 0, the contents is returned
    unchanged.

    If max_consecutive_newlines == 0, all strings of consecutive
    newlines are replaced by a single space.

    If max_consecutive_newlines > 0, all strings of consecutive
    newlines in excess of max_consecutive_newlines are replaced by
    max_consecutive_newlines newlines.
    """
    if max_consecutive_newlines < 0:
        return contents
    elif max_consecutive_newlines == 0:
        return re.sub(r'\n+', ' ', contents)
    else:
        newlines = r'\n' * max_consecutive_newlines
        return re.sub(newlines + '+', newlines, contents)



from .custom_arguments import DEFAULT_LOG

__all__ = ["process_maybe_list"]

def process_maybe_list(ls, log=DEFAULT_LOG):
    if ls is None: return tuple()
    if isinstance(ls, str): return tuple([ls])
    if isinstance(ls, tuple): return ls
    if isinstance(ls, list): return tuple(ls)
    log("Unknown type '%s' of list '%s'" % (str(type(ls)), repr(ls)))
    return tuple(ls)



import re

__all__ = ["transform_abstract_to_admit"]

def DEFAULT_LOG(text, level=1):
    if level <= 1: print(text)

TERM_CHAR = r"[\w']"
ABSTRACT_NO_PARENS_DOT = re.compile(r"(^\s*|\.\s+|;\s*)abstract\s+(?:[^\(\);\.]|\.%s)+(?=\.\s|\.$)" % TERM_CHAR, re.MULTILINE)

def transform_abstract_to_admit_statement(statement, agressive=False, log=DEFAULT_LOG):
    # remove the unparenthesized ones
    statement = ABSTRACT_NO_PARENS_DOT.sub(r'\1admit', statement)

    if 'abstract' not in statement:
        return statement

    # now look at the parenthesized abstracts
    ready_for_abstract = True
    in_abstract = False
    abstract_paren_level = 0
    rtn = []
    cur = []
    for term in re.split('([;\.\(\)])', statement):
        log('in_abstract: %d; abstract_paren_level: %d; agressive: %d; ready_for_abstract: %d;\n^term: %s' %
            (in_abstract, abstract_paren_level, agressive, ready_for_abstract, term),
            level=3)
        if in_abstract:
            if abstract_paren_level == 0 and term in tuple(';.'):
                if term == ';':
                    if agressive:
                        rtn.append(' admit;')
                        log("Appending ' admit;' to rtn", level=3)
                    else:
                        cur.append(term)
                        rtn.append(''.join(cur))
                        log("Appending %s to rtn" % repr(''.join(cur)), level=3)
                else:
                    log("Appending ' admit.' to rtn", level=3)
                    rtn.append(' admit.')
                in_abstract = False
                cur = []
                log("Setting in_abstract to false and emptying cur", level=3)
            elif abstract_paren_level < 0:
                log('Warning: abstract_paren_level messed up on statement %s' % repr(statement))
                in_abstract = False
                cur.append(term)
                rtn.append(''.join(cur))
                cur = []
                abstract_paren_level = 0
                log("Setting in_abstract to false, abstract_paren_level to 0, emptying cur, and\nappending %s to rtn" %
                    repr(''.join(cur)), level=3)
            else:
                if term == '(':
                    abstract_paren_level += 1
                elif term == ')':
                    abstract_paren_level -= 1
                cur.append(term)
                log("Setting abstract_paren_level to %d and\nappending %s to cur" % (abstract_paren_level, repr(term)), level=3)
        else:
            if ready_for_abstract and term.strip() == 'abstract':
                cur.append(term)
                in_abstract = True
                log("Found %s (appending to cur), now in_abstract" % repr(term), level=3)
                if abstract_paren_level != 0:
                    log('Warning: abstract_paren_level messed up before abstract on statement %s' % repr(statement))
                    abstract_paren_level = 0
            else:
                if term in tuple('.;'):
                    log("Found %s (appending to rtn), ready for abstract" % repr(term), level=3)
                    ready_for_abstract = True
                elif term.strip():
                    log("Found %s (appending to rtn), not ready for abstract" % repr(term), level=3)
                    ready_for_abstract = False
                rtn.append(term)
    log("Done.  Appending %s to rtn." % repr(''.join(cur)), level=3)
    rtn.append(''.join(cur))

    return ''.join(rtn)

def transform_abstract_to_admit(cur_definition, rest_definitions, **kwargs):
    # shallow copy
    cur_definition = dict(cur_definition)
    cur_definition['statements'] = tuple(transform_abstract_to_admit_statement(i, **kwargs)
                                         for i in cur_definition['statements'])
    cur_definition['statement'] = '\n'.join(cur_definition['statements'])
    return cur_definition



import re, time
from subprocess import Popen, PIPE, STDOUT
from . import split_definitions_old
from .split_file import postprocess_split_proof_term
from .coq_version import get_coq_accepts_time, get_coq_accepts_emacs
from .coq_running_support import get_proof_term_works_with_time
from .custom_arguments import DEFAULT_LOG, LOG_ALWAYS
from . import util

__all__ = ["join_definitions", "split_statements_to_definitions",
           "PREFER_PASSING", "PASSING", "PREFER_NONPASSING", "NONPASSING",
           "get_preferred_passing", "get_fallback_passing"]

PREFER_PASSING, PASSING, PREFER_NONPASSING, NONPASSING = \
    'prefer-passing', 'passing', 'prefer-nonpassing', 'nonpassing'

DEFAULT_PREFERENCE_KEY = 'parse_with'
def get_preferred_fallback_passing(key, preference_key=DEFAULT_PREFERENCE_KEY, **env):
    if env[preference_key] in (PREFER_PASSING, PASSING, PREFER_NONPASSING, NONPASSING):
        default = env.get('passing_' + key) if env[preference_key] in (PREFER_PASSING, PASSING) else env.get(key)
        fallback = None if env[preference_key] in (PASSING, NONPASSING) \
            else env.get('passing_' + key) if env[preference_key] == PREFER_NONPASSING else env.get(key)
        return (default, fallback)
    raise ValueError('Invalid value for %s: must be one of %s' % (preference_key, ', '.join(map(repr, (PREFER_PASSING, PASSING, PREFER_NONPASSING, NONPASSING)))))

def get_preferred_passing(key, **env):
    default, fallback = get_preferred_fallback_passing(key, **env)
    return default if default is not None else fallback

def get_fallback_passing(key, **env):
    default, fallback = get_preferred_fallback_passing(key, **env)
    # no fallback if we use fallback as default
    return fallback if default is not None else None

def get_definitions_diff(previous_definition_string, new_definition_string):
    """Returns a triple of lists (definitions_removed,
    definitions_shared, definitions_added)"""
    old_definitions = [i for i in previous_definition_string.split('|') if i]
    new_definitions = [i for i in new_definition_string.split('|') if i]
    if all(i == 'branch' for i in old_definitions + new_definitions): # work
        # around bug #5577 when all theorem names are "branch", we
        # don't assume that names are unique, and instead go by
        # ordering
        removed = []
        shared = []
        added = []
        for i in range(max((len(old_definitions), len(new_definitions)))):
            if i < len(old_definitions) and i < len(new_definitions):
                if old_definitions[i] == new_definitions[i]:
                    shared.append(old_definitions[i])
                else:
                    removed.append(old_definitions[i])
                    added.append(new_definitions[i])
            elif i < len(old_definitions):
                removed.append(old_definitions[i])
            elif i < len(new_definitions):
                added.append(new_definitions[i])
        return (tuple(removed), tuple(shared), tuple(added))
    else:
        return (tuple(i for i in old_definitions if i not in new_definitions),
                tuple(i for i in old_definitions if i in new_definitions),
                tuple(i for i in new_definitions if i not in old_definitions))

def strip_newlines(string):
    if not string: return string
    if string[0] == '\n': return string[1:]
    if string[-1] == '\n': return string[:-1]
    return string

def split_statements_to_definitions(statements, log=DEFAULT_LOG, coqtop='coqtop', coqtop_args=tuple(), fallback_coqtop=None, fallback_coqtop_args=None, preference_key=DEFAULT_PREFERENCE_KEY, **kwargs):
    """Splits a list of statements into chunks which make up
    independent definitions/hints/etc."""
    if preference_key is not None:
        env = {'coqtop': coqtop, 'coqtop_args': coqtop_args, 'log': log}
        env.update(kwargs)
        coqtop, fallback_coqtop = get_preferred_fallback_passing('coqtop', preference_key=preference_key, **env)
        coqtop_args, fallback_coqtop_args = get_preferred_fallback_passing('coqtop_args', preference_key=preference_key, **env)
        return split_statements_to_definitions(statements, log=log, coqtop=coqtop, coqtop_args=coqtop_args, fallback_coqtop=fallback_coqtop, fallback_coqtop_args=fallback_coqtop_args, preference_key=None, **kwargs)
    if coqtop is None and fallback_coqtop is not None:
        return split_statements_to_definitions(statements, log=log, coqtop=fallback_coqtop, coqtop_args=fallback_coqtop_args, fallback_coqtop=None, fallback_coqtop_args=None, preference_key=None, **kwargs)
    def fallback():
        if fallback_coqtop is not None:
            log("Your preferred version of coqtop (%s) doesn't support -emacs -time.  Falling back to fallback coqtop (%s)." % (coqtop, fallback_coqtop))
            return split_statements_to_definitions(statements, log=log, coqtop=fallback_coqtop, coqtop_args=fallback_coqtop_args, fallback_coqtop=None, fallback_coqtop_args=None, preference_key=None, **kwargs)
        else:
            log("Your version of coqtop doesn't support -emacs -time.  Falling back to more error-prone method.")
            return split_definitions_old.split_statements_to_definitions(statements, log=log, coqtop=coqtop, coqtop_args=coqtop_args)
    # check for -time
    if not get_coq_accepts_time(coqtop, log=log) or not get_coq_accepts_emacs(coqtop, log=log):
        return fallback()
    if not get_proof_term_works_with_time(coqtop, is_coqtop=True, log=log, **kwargs):
        statements = postprocess_split_proof_term(statements, log=log, **kwargs)
    p = Popen([coqtop, '-q', '-emacs', '-time'] + list(coqtop_args), stdout=PIPE, stderr=STDOUT, stdin=PIPE)
    chars_time_reg = re.compile(r'Chars ([0-9]+) - ([0-9]+) [^\s]+'.replace(' ', r'\s*'))
    prompt_reg = re.compile(r'^(.*)<prompt>([^<]*?) < ([0-9]+) ([^<]*?) ([0-9]+) < ([^<]*)$'.replace(' ', r'\s*'),
                            flags=re.DOTALL)
    defined_reg = re.compile(r'^(?:<infomsg>)?([^\s]+) is (?:defined|assumed|declared)(?:</infomsg>)?$', re.MULTILINE)
    # goals and definitions are on stdout, prompts are on stderr
    statements_string = '\n'.join(statements) + '\n\n'
    statements_bytes = statements_string.encode('utf-8')
    log('Sending statements to coqtop...')
    log(statements_string, level=3)
    (stdout, stderr) = p.communicate(input=statements_bytes)
    stdout = util.s(stdout)
    if 'know what to do with -time' in stdout.strip().split('\n')[0]:
        # we're using a version of coqtop that doesn't support -time
        return fallback()
    log('Done.  Splitting to definitions...')

    rtn = []
    cur_definition = {}
    last_definitions = '||'
    cur_definition_names = '||'
    last_char_end = 0

    #log('re.findall(' + repr(r'Chars ([0-9]+) - ([0-9]+) [^\s]+ (.*?)<prompt>([^<]*?) < ([0-9]+) ([^<]*?) ([0-9]+) < ([^<]*?)</prompt>'.replace(' ', r'\s*')) + ', ' + repr(stdout) + ', ' + 'flags=re.DOTALL)', level=3)
    for i, prompt in enumerate(stdout.split('</prompt>')):
        log('Processing:\n%s' % prompt, level=4)
        if prompt.strip() == '': continue
        times = chars_time_reg.findall(prompt)
        if len(times) == 0:
            if i == 0:
                log('Warning: ignoring %s' % repr(prompt), level=3)
            else:
                log('Warning: Could not find timing info in %s' % repr(prompt), level=1)
            continue
        elif len(times) > 1:
            log('Warning: found multiple timing info in %s: %s' % (repr(prompt), repr(times)), level=1)
        full_response_text = chars_time_reg.sub(' ', prompt)
        for char_start, char_end in times:
            char_start, char_end = int(char_start), int(char_end)
            # if we've travelled backwards in time, as in
            # COQBUG(https://github.com/coq/coq/issues/14475); we just
            # ignore this statement
            if char_end <= last_char_end: continue
            match = prompt_reg.match(full_response_text)
            if not match:
                log('Warning: Could not find statements in %d:%d: %s' % (char_start, char_end, full_response_text), level=LOG_ALWAYS)
                continue
            response_text, cur_name, line_num1, cur_definition_names, line_num2, unknown = match.groups()
            statement = strip_newlines(statements_bytes[last_char_end:char_end].decode('utf-8'))
            last_char_end = char_end

            terms_defined = defined_reg.findall(response_text.strip())

            definitions_removed, definitions_shared, definitions_added = get_definitions_diff(last_definitions, cur_definition_names)

            # first, to be on the safe side, we add the new
            # definitions key to the dict, if it wasn't already there.
            if cur_definition_names.strip('|') and cur_definition_names not in cur_definition:
                cur_definition[cur_definition_names] = {'statements':[], 'terms_defined':[]}


            log((statement, (char_start, char_end), definitions_removed, terms_defined, 'last_definitions:', last_definitions, 'cur_definition_names:', cur_definition_names, cur_definition.get(last_definitions, []), cur_definition.get(cur_definition_names, []), response_text), level=2)


            # first, we handle the case where we have just finished
            # defining something.  This should correspond to
            # len(definitions_removed) > 0 and len(terms_defined) > 0.
            # If only len(definitions_removed) > 0, then we have
            # aborted something (or we're in Coq >= 8.11; see
            # https://github.com/coq/coq/issues/15201).  If only
            # len(terms_defined) > 0, then we have defined something
            # with a one-liner.
            if definitions_removed:
                cur_definition[last_definitions]['statements'].append(statement)
                cur_definition[last_definitions]['terms_defined'] += terms_defined
                if cur_definition_names.strip('|'):
                    # we are still inside a definition.  For now, we
                    # flatten all definitions.
                    #
                    # TODO(jgross): Come up with a better story for
                    # nested definitions.
                    cur_definition[cur_definition_names]['statements'] += cur_definition[last_definitions]['statements']
                    cur_definition[cur_definition_names]['terms_defined'] += cur_definition[last_definitions]['terms_defined']
                    del cur_definition[last_definitions]
                else:
                    # we're at top-level, so add this as a new
                    # definition
                    rtn.append({'statements':tuple(cur_definition[last_definitions]['statements']),
                                'statement':'\n'.join(cur_definition[last_definitions]['statements']),
                                'terms_defined':tuple(cur_definition[last_definitions]['terms_defined'])})
                    del cur_definition[last_definitions]
                    # print('Adding:')
                    # print(rtn[-1])
            elif terms_defined:
                if cur_definition_names.strip('|'):
                    # we are still inside a definition.  For now, we
                    # flatten all definitions.
                    #
                    # TODO(jgross): Come up with a better story for
                    # nested definitions.
                    cur_definition[cur_definition_names]['statements'].append(statement)
                    cur_definition[cur_definition_names]['terms_defined'] += terms_defined
                else:
                    # we're at top level, so add this as a new
                    # definition
                    rtn.append({'statements':(statement,),
                                'statement':statement,
                                'terms_defined':tuple(terms_defined)})

            # now we handle the case where we have just opened a fresh
            # definition.  We've already added the key to the
            # dictionary.
            elif definitions_added:
                # print(definitions_added)
                cur_definition[cur_definition_names]['statements'].append(statement)
            else:
                # if we're in a definition, append the statement to
                # the queue, otherwise, just add it as it's own
                # statement
                if cur_definition_names.strip('|'):
                    cur_definition[cur_definition_names]['statements'].append(statement)
                else:
                    rtn.append({'statements':(statement,),
                                'statement':statement,
                                'terms_defined':tuple()})

            last_definitions = cur_definition_names

    log((last_definitions, cur_definition_names), level=2)
    if last_definitions.strip('||'):
        rtn.append({'statements':tuple(cur_definition[cur_definition_names]['statements']),
                    'statement':'\n'.join(cur_definition[cur_definition_names]['statements']),
                    'terms_defined':tuple(cur_definition[cur_definition_names]['terms_defined'])})
        del cur_definition[last_definitions]

    if last_char_end + 1 < len(statements_bytes):
        last_statement = statements_bytes[last_char_end:].decode('utf-8')
        log('Appending end of code from %d to %d: %s' % (last_char_end, len(statements_bytes), last_statement), level=2)
        last_statement = strip_newlines(last_statement)
        rtn.append({'statements':tuple(last_statement,),
                    'statement':last_statement,
                    'terms_defined':tuple()})

    return rtn

def join_definitions(definitions):
    return '\n'.join(i['statement'] for i in definitions)



import os, time
from .memoize import memoize
from . import util

__all__ = ["clean_v_file", "clean_extra_coq_files", "write_bytes_to_file", "read_bytes_from_file", "write_to_file", "read_from_file", "restore_file"]

def clean_extra_coq_files(v_file_name, extra_exts=tuple()):
    for pre in ('', '.'):
        for ext in tuple(list(extra_exts) + ['.glob', '.vo', '.d', '.v.d', '.aux', '.vos', '.vok']):
            name = ''.join((os.path.dirname(v_file_name[:-2]), os.sep, pre, os.path.basename(v_file_name[:-2]), ext))
            if os.path.exists(name):
                os.remove(name)

def clean_v_file(file_name):
    clean_extra_coq_files(file_name, extra_exts=('.v',))

def backup(file_name, ext='.bak'):
    if not ext:
        raise ValueError
    if os.path.exists(file_name):
        backup(file_name + ext, ext=ext)
        os.rename(file_name, file_name + ext)

@memoize
def memobackup(file_name, ext='.bak'):
    return backup(file_name, ext)

FILE_CACHE = {}

def write_bytes_to_file(file_name, contents, do_backup=False, backup_ext='.bak', memoize=False):
    assert(contents is bytes(contents))
    backed_up = False
    while not backed_up:
        try:
            if do_backup:
                if memoize:
                    memobackup(file_name, ext=backup_ext)
                else:
                    backup(file_name, ext=backup_ext)
            backed_up = True
        except IOError as e:
            print('Warning: f.write(%s) failed with %s\nTrying again in 10s' % (file_name, repr(e)))
            time.sleep(10)
    written = False
    while not written:
        try:
            with open(file_name, 'wb') as f:
                f.write(contents)
            written = True
            FILE_CACHE[file_name] = (os.stat(file_name).st_mtime, contents)
        except IOError as e:
            print('Warning: f.write(%s) failed with %s\nTrying again in 10s' % (file_name, repr(e)))
            time.sleep(10)

def write_to_file(file_name, contents, *args, **kwargs):
    return write_bytes_to_file(file_name, contents.replace('\n', os.linesep).encode('utf-8'), *args, **kwargs)

def restore_file(file_name, backup_ext='.bak', backup_backup_ext='.unbak'):
    if not os.path.exists(file_name + backup_ext):
        raise IOError
    if os.path.exists(file_name):
        if backup_backup_ext:
            backup(file_name, backup_backup_ext)
        else:
            os.remove(file_name)
    os.rename(file_name + backup_ext, file_name)

def read_bytes_from_file(file_name):
    if file_name in FILE_CACHE.keys() and os.stat(file_name).st_mtime == FILE_CACHE[file_name][0]:
        return FILE_CACHE[file_name][1]
    with open(file_name, 'rb') as f:
        FILE_CACHE[file_name] = (os.stat(file_name).st_mtime, f.read())
        return FILE_CACHE[file_name][1]

def read_from_file(file_name):
    return util.normalize_newlines(read_bytes_from_file(file_name).decode('utf-8'))



__all__ = [
    'absolutize-imports',
    'find_bug',
    'get-admitted-names',
    'inline_imports',
    'minimize_requires',
    'move-requires',
    'move-vernaculars',
    'proof-using-helper',
  ]



__all__ = ['strip_comments']

def strip_comments(contents):
    """Strips the comments from coq code in contents.

    The strings in contents are only preserved if there are no
    comment-like tokens inside of strings.  Stripping should be
    successful and correct, regardless of whether or not there are
    comment-like tokens in strings.

    The behavior of this method is undefined if there are any
    notations which change the meaning of '(*', '*)', or '"'.

    Note that we take some extra care to leave *) untouched when it
    does not terminate a comment.
    """
    contents = contents.replace('(*', ' (* ').replace('*)', ' *) ')
    tokens = contents.split(' ')
    rtn = []
    is_string = False
    comment_level = 0
    for token in tokens:
        do_append = (comment_level == 0)
        if is_string:
            if token.count('"') % 2 == 1: # there are an odd number of '"' characters, indicating that we've ended the string
                is_string = False
        elif token.count('"') % 2 == 1: # there are an odd number of '"' characters, so we're starting a string
            is_string = True
        elif token == '(*':
            comment_level += 1
            do_append = False
        elif comment_level > 0 and token == '*)':
            comment_level -= 1
        if do_append:
            rtn.append(token)
    return ' '.join(rtn).replace(' (* ', '(*').replace(' *) ', '*)').strip('\n\t ')



#!/usr/bin/env python
import shutil, os, os.path, sys
from .argparse_compat import argparse
from .import_util import IMPORT_ABSOLUTIZE_TUPLE, ALL_ABSOLUTIZE_TUPLE
from .custom_arguments import add_libname_arguments, update_env_with_libnames, update_env_with_coqpath_folders, add_logging_arguments, process_logging_arguments
from .coq_version import get_coqc_coqlib, DEFAULT_COQTOP
from .replace_imports import include_imports

__all__ = ['main']

parser = argparse.ArgumentParser(description='Inline the imports of a file')
parser.add_argument('input_file', metavar='IN_FILE', type=argparse.FileType('r'),
                    help='a .v file to inline the imports of')
parser.add_argument('output_file', metavar='OUT_FILE', type=argparse.FileType('w'),
                    help='a .v file to write to')
parser.add_argument('--fast-merge-imports', dest='fast_merge_imports',
                    action='store_const', const=True, default=False,
                    help='Use a faster method for combining imports')
parser.add_argument('--no-deps', dest='use_coq_makefile_for_deps',
                    action='store_const', const=False, default=True,
                    help=("Don't do dependency analysis with coq_makefile."))
parser.add_argument('--no-pwd-deps', dest='walk_tree',
                    action='store_const', const=False, default=True,
                    help=("Don't add all files in the current directory to the dependency analysis."))
parser.add_argument('--no-wrap-modules', dest='wrap_modules',
                    action='store_const', const=False, default=True,
                    help=("Don't wrap imports in Modules.  By default, the " +
                          "contents of each file is wrapped in its own " +
                          "module to deal with renaming issues.  This " +
                          "can cause issues with subdirectories."))
parser.add_argument('--absolutize-constants', dest='absolutize',
                    action='store_const', default=IMPORT_ABSOLUTIZE_TUPLE, const=ALL_ABSOLUTIZE_TUPLE,
                    help=("Replace constants with fully qualified versions.  " +
                          "By default, all constants are not fully qualified.  If you have " +
                          "many overlapping file names in different directories " +
                          "and use partially qualified names that differ depending " +
                          "on which files have been Required, not absolutizing constants " +
                          "may cause name resolution to fail."))
parser.add_argument('--coqbin', metavar='COQBIN', dest='coqbin', type=str, default='',
                    help='The path to a folder containing the coqc and coqtop programs.')
parser.add_argument('--coqc', metavar='COQC', dest='coqc', type=str, default='coqc',
                    help='The path to the coqc program.')
parser.add_argument('--coqtop', metavar='COQTOP', dest='coqtop', type=str, default=DEFAULT_COQTOP,
                    help=('The path to the coqtop program (default: %s).' % DEFAULT_COQTOP))
parser.add_argument('--coqc-args', metavar='ARG', dest='coqc_args', type=str, nargs='?',
                    help='Arguments to pass to coqc; e.g., " -indices-matter" (leading and trailing spaces are stripped)')
parser.add_argument('--coqtop-args', metavar='ARG', dest='coqtop_args', type=str, nargs='?',
                    help='Arguments to pass to coqtop; e.g., " -indices-matter" (leading and trailing spaces are stripped)')
parser.add_argument('--coq_makefile', metavar='COQ_MAKEFILE', dest='coq_makefile', type=str, default='coq_makefile',
                    help='The path to the coq_makefile program.')
parser.add_argument('--inline-user-contrib', dest='inline_user_contrib',
                    action='store_const', const=True, default=False,
                    help=("Attempt to inline requires from the user-contrib folder"))
add_libname_arguments(parser)
add_logging_arguments(parser)

def main():
    args = process_logging_arguments(parser.parse_args())

    env = {
        'log': args.log,
        'coqc': (args.coqc if args.coqbin == '' else os.path.join(args.coqbin, args.coqc)),
        'absolutize': args.absolutize,
        'as_modules': args.wrap_modules,
        'fast': args.fast_merge_imports,
        'coq_makefile': args.coq_makefile,
        'use_coq_makefile_for_deps': args.use_coq_makefile_for_deps,
        'walk_tree': args.walk_tree,
        }
    update_env_with_libnames(env, args)
    if args.inline_user_contrib: update_env_with_coqpath_folders(env, os.path.join(get_coqc_coqlib(env['coqc'], coq_args=env['coqc_args'], **env), 'user-contrib'))

    filename = args.input_file.name
    args.input_file.close()

    rtn = include_imports(filename, **env)

    args.output_file.write(rtn)
    args.output_file.close()

if __name__ == '__main__':
    main()



import subprocess, re
from .strip_comments import strip_comments
from .custom_arguments import DEFAULT_LOG
from .coq_version import get_coq_accepts_time
from . import util
from .util import PY3
if PY3: from .util import raw_input

__all__ = ["split_coq_file_contents", "split_coq_file_contents_with_comments", "get_coq_statement_byte_ranges", "UnsupportedCoqVersionError", "postprocess_split_proof_term", "split_leading_comments_and_whitespace"]

def fill_kwargs(kwargs):
    ret = {
        'log'                   : DEFAULT_LOG,
        'coqc'                  : 'coqc',
        'coqc_args'             : tuple(),
    }
    ret.update(kwargs)
    return ret

class UnsupportedCoqVersionError(Exception):
    pass

def merge_quotations(statements, sp=' '):
    """If there are an odd number of "s in a statement, assume that we
    broke the middle of a string.  We recombine that string."""

    cur = None
    for i in statements:
        if i.count('"') % 2 != 0:
            if cur is None:
                cur = i
            else:
                yield (cur + sp + i)
                cur = None
        elif cur is None:
            yield i
        else:
            cur += sp + i

BRACES = '{}-*+'

def split_leading_braces(statement):
    while statement.strip() and statement.strip()[0] in BRACES:
        idx = statement.index(statement.strip()[0]) + 1
        first, rest = statement[:idx], statement[idx:]
        #if 'not_reachable_iff' in statement or 'not_reachable_iff' in first:
        #    print((first, rest))
        yield first
        statement = rest
    if statement:
        yield statement

def split_leading_comments_and_whitespace(text):
    comment_level = 0
    in_str = False
    skip = False
    for i, (ch, next_ch) in enumerate(zip(text, text[1:] + '\0')):
        if skip:
            skip = False
        elif in_str:
            if ch in '"': in_str = False
        elif ch == '(' and next_ch == '*':
            comment_level += 1
            skip = True
        elif ch == '*' and next_ch == ')':
            comment_level -= 1
            skip = True
        elif comment_level > 0:
            if ch in '"': in_str = True
        elif ch.isspace():
            pass
        else:
            return text[:i], text[i:]
    return text, ''

def split_merge_comments(statements):
    """Find open and close comments."""

    def is_genuine_ending(cur):
        return ((cur.strip()[:2] == '(*' and cur.strip()[-2:] == '*)') # cur is just a comment
                or cur.strip()[-1:] in '.}'
                or cur.strip() in BRACES)

    cur = None
    comment_level = 0
    for stmt in statements:
        str_count = 0
        for i in re.split(r'(?<=\*\)) ', stmt.replace('*)', '*) ')):
            if str_count % 2 == 0:
                i2 = re.sub('"[^"]*"|"[^"]*$', '', i)
            else:
                i2 = re.sub('^[^"]*"|"[^"]*"', '', i)
            str_count += i.count('"')
            #print((repr(i), comment_level, repr(i2)))
            if '(*' not in i2 and '*)' not in i2:
                if comment_level < 0:
                    if cur is not None:
                        for ret in split_leading_braces(cur + i):
                            yield ret
                        cur = None
                    else:
                        raw_input('UNEXPECTED COMMENT: %s' % i)
                        yield i
                    comment_level = 0
                elif comment_level == 0:
                    if cur is None:
                        for ret in split_leading_braces(i):
                            yield ret
                    else:
                        cur += i
                elif cur is None:
                    cur = i
                else:
                    cur += i
            else:
                comment_level += i2.count('(*') - i2.count('*)')
                if cur is None:
                    cur = i
                else:
                    cur += i
            if cur is not None:
                curs = list(split_leading_braces(cur))
                while curs and curs[0].strip() in BRACES:
                    yield curs.pop(0)
                cur = ''.join(curs) if curs else None
            if cur is not None and is_genuine_ending(cur) and comment_level == 0: # clear out cur, if we're done with it
                yield cur
                cur = None
    if cur is not None:
        print('Unterminated comment')
        yield cur


def split_coq_file_contents(contents):
    """Splits the contents of a coq file into multiple statements.

    This is done by finding one or three periods followed by
    whitespace.  This is a dumb algorithm, but it seems to be (nearly)
    the one that ProofGeneral and CoqIDE use.

    We additionally merge lines inside of quotations."""
    return list(merge_quotations(re.split('(?<=[^\.]\.\.\.)\s|(?<=[^\.]\.)\s', strip_comments(contents))))

def split_coq_file_contents_with_comments(contents):
    statements = re.split(r'(?<=[^\.]\.\.\.) |(?<=[^\.]\.) ',
                          re.sub(r'((?<=[^\.]\.\.\.)\s|(?<=[^\.]\.)\s)', r' \1', contents))
    #if contents != ''.join(statements):
    #    print('Splitting failed (initial split)!')
    #qstatements = list(merge_quotations(statements, sp=''))
    #if contents != ''.join(qstatements):
    #    print('Splitting failed (quote merge)!')
    #cstatements = list(split_merge_comments(qstatements))
    #if contents != ''.join(cstatements):
    #    print('Splitting failed (comment merge)!')
    return list(split_merge_comments(merge_quotations(statements, sp='')))

PROOF_TERM_REG = re.compile(r'(\s*)Proof(\s+)(.*?)\.(\s*)', flags=re.DOTALL | re.MULTILINE)
def postprocess_split_proof_term_iter(statements, on_first_example=None):
    """Returns an iterator for the statements passed which changes [Proof
    term.] into [Proof. exact (term). Qed.] unless "term" begins with
    "with" or "using"."""
    for statement_num, statement in enumerate(statements):
        match = PROOF_TERM_REG.match(statement)
        if match:
            indent, after_proof_space, proof_term, post_space = match.groups()
            if re.split(r'\s', proof_term)[0] not in ('using', 'with'): # Proof using ... and Proof with ... are special
                if on_first_example:
                    on_first_example(statement_num, statement, proof_term)
                    on_first_example = None
                yield indent + 'Proof.'
                yield after_proof_space + 'exact (' + proof_term + ').'
                yield ' Qed.' + post_space
            else:
                yield statement
        else:
            yield statement
def postprocess_split_proof_term(statements, do_warn=True, **kwargs):
    def do_warn_method(statement_num, statement, proof_term):
        kwargs['log']("""Warning: Your version of Coq suffers from bug #5349 (https://coq.inria.fr/bugs/show_bug.cgi?id=5349)
and does not support [Proof (term).] with -time.  Falling back to
replacing [Proof (term).] with [Proof. exact (term). Qed.], which may fail.""")
    if do_warn and 'log' in kwargs.keys():
        do_warn = do_warn_method
    else:
        do_warn = None
    return list(postprocess_split_proof_term_iter(statements, do_warn))

RANGE_REG = re.compile(r'Chars ([0-9]+) - ([0-9]+) [^\s]+', flags=re.DOTALL)

def get_coq_statement_byte_ranges(file_name, coqc, **kwargs):
    kwargs = fill_kwargs(kwargs)
    if not get_coq_accepts_time(coqc, **kwargs):
        raise UnsupportedCoqVersionError

    p = subprocess.Popen([coqc, '-q', '-time'] + list(kwargs['coqc_args']) + [file_name], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, stdin=subprocess.PIPE)
    (stdout, stderr) = p.communicate()

    ranges = tuple((int(start), int(finish)) for start, finish in RANGE_REG.findall(util.s(stdout)))
    return ranges



from __future__ import with_statement, print_function
import os, sys, tempfile, subprocess, re, time, math, glob, threading, shutil, traceback
from .Popen_noblock import Popen_async, Empty
from .memoize import memoize
from .file_util import clean_v_file
from .util import re_escape
from .custom_arguments import LOG_ALWAYS
from .coq_version import group_coq_args, get_coqc_help
from . import util

__all__ = ["has_error", "get_error_line_number", "get_error_byte_locations", "make_reg_string", "get_coq_output", "get_error_string", "get_timeout", "reset_timeout", "reset_coq_output_cache", "is_timeout"]

ACTUAL_ERROR_REG_STRING = '(?!Warning)(?!The command has indeed failed with message:)' # maybe we should just require Error or Timeout?
DEFAULT_PRE_PRE_ERROR_REG_STRING = 'File "[^"]+", line ([0-9]+), characters [0-9-]+:\n'
DEFAULT_PRE_ERROR_REG_STRING = DEFAULT_PRE_PRE_ERROR_REG_STRING + ACTUAL_ERROR_REG_STRING
DEFAULT_PRE_ERROR_REG_STRING_WITH_BYTES = 'File "[^"]+", line ([0-9]+), characters ([0-9]+)-([0-9]+):\n' + ACTUAL_ERROR_REG_STRING
DEFAULT_ERROR_REG_STRING = DEFAULT_PRE_ERROR_REG_STRING + '((?:.|\n)+)'
DEFAULT_ERROR_REG_STRING_WITH_BYTES = DEFAULT_PRE_ERROR_REG_STRING_WITH_BYTES + '((?:.|\n)+)'
DEFAULT_ERROR_REG_STRING_GENERIC = DEFAULT_PRE_PRE_ERROR_REG_STRING + '(%s)'

def clean_output(output):
    return util.normalize_newlines(output)

@memoize
def get_coq_accepts_fine_grained_debug(coqc, debug_kind):
    temp_file = tempfile.NamedTemporaryFile(suffix='.v', dir='.', delete=True)
    temp_file_name = temp_file.name
    p = subprocess.Popen([coqc, "-q", "-d", debug_kind, temp_file_name], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    temp_file.close()
    clean_v_file(temp_file_name)
    return 'Unknown option -d' not in util.s(stdout) and '-d: no such file or directory' not in util.s(stdout) and 'There is no debug flag' not in util.s(stdout)

def get_coq_debug_native_compiler_args(coqc):
    if get_coq_accepts_fine_grained_debug(coqc, "native-compiler"): return ["-d", "native-compiler"]
    return ["-debug"]

@memoize
def get_error_match(output, reg_string=DEFAULT_ERROR_REG_STRING, pre_reg_string=DEFAULT_PRE_ERROR_REG_STRING):
    """Returns the final match of reg_string"""
    locations = [0] + [m.start() for m in re.finditer(pre_reg_string, output)]
    reg = re.compile(reg_string)
    results = (reg.search(output[start_loc:]) for start_loc in reversed(locations))
    for result in results:
        if result: return result
    return None

@memoize
def has_error(output, reg_string=DEFAULT_ERROR_REG_STRING, pre_reg_string=DEFAULT_PRE_ERROR_REG_STRING):
    """Returns True if the coq output encoded in output has an error
    matching the given regular expression, False otherwise.
    """
    errors = get_error_match(output, reg_string=reg_string, pre_reg_string=pre_reg_string)
    if errors:
        return True
    else:
        return False

TIMEOUT_POSTFIX = '\nTimeout! (external)'

@memoize
def is_timeout(output):
    """Returns True if the output was killed with a timeout, False otherwise"""
    return output.endswith(TIMEOUT_POSTFIX)

@memoize
def get_error_line_number(output, reg_string=DEFAULT_ERROR_REG_STRING, pre_reg_string=DEFAULT_PRE_ERROR_REG_STRING):
    """Returns the line number that the error matching reg_string
    occured on.

    Precondition: has_error(output, reg_string)
    """
    errors = get_error_match(output, reg_string=reg_string, pre_reg_string=pre_reg_string)
    return int(errors.groups()[0])

@memoize
def get_error_byte_locations(output, reg_string=DEFAULT_ERROR_REG_STRING_WITH_BYTES, pre_reg_string=DEFAULT_PRE_ERROR_REG_STRING_WITH_BYTES):
    """Returns the byte locations that the error matching reg_string
    occured on.

    Precondition: has_error(output, reg_string)
    """
    errors = get_error_match(output, reg_string=reg_string, pre_reg_string=pre_reg_string)
    return (int(errors.groups()[1]), int(errors.groups()[2]))

@memoize
def get_error_string(output, reg_string=DEFAULT_ERROR_REG_STRING, pre_reg_string=DEFAULT_PRE_ERROR_REG_STRING):
    """Returns the error string of the error matching reg_string.

    Precondition: has_error(output, reg_string)
    """
    errors = get_error_match(output, reg_string=reg_string, pre_reg_string=pre_reg_string)
    return errors.groups()[1]

@memoize
def make_reg_string(output, strict_whitespace=False):
    """Returns a regular expression for matching the particular error
    in output.

    Precondition: has_error(output)
    """
    unstrictify_whitespace = (lambda s: s)
    if not strict_whitespace:
        unstrictify_whitespace = (lambda s: re.sub(r'(?:\\s\\+)+', r'\\s+', re.sub(r'(?:\\ )+', r'\\s+', re.sub(r'(\\n|\n)(?:\\ )+', r'\\s+', s.replace('\\\n', '\n')))).replace('\n', '\s').replace(r'\s+\s', r'\s+').replace(r'\s++', r'\s+'))

    error_string = get_error_string(output).strip()
    if not util.PY3: error_string = error_string.decode('utf-8')
    if 'Anomaly' in error_string and re.search(r'Constant [^\s]+\s+does not appear in the environment', error_string):
        re_string = re.sub(r'(Constant\\ )[^\s]+(\\ )',
                           r'\1[^\\s]+\2',
                           re_escape(error_string))
    elif 'Anomaly' in error_string and re.search(r'Universe [^\s]+ undefined', error_string):
        re_string = re.sub(r'(Universe\\ )[^\s]+(\\ )',
                           r'\1[^\\s]+\2',
                           re_escape(error_string))
    elif 'Universe inconsistency' in error_string or 'universe inconsistency' in error_string:
        re_string = re.sub(r'([Uu]niverse\\ inconsistency.*) because(.|\n)*',
                           r'\1 because.*',
                           re_escape(error_string))
        re_string = re.sub(r'(\s)[^\s]+?\.([0-9]+)',
                           r'\1[^\\s]+?\\.\2',
                           re_string)
    elif 'Unsatisfied constraints' in error_string:
        re_string = re.sub(r'Error:\\ Unsatisfied\\ constraints:.*(?:\n.+)*.*\\\(maybe\\ a\\ bugged\\ tactic\\\)',
                           r'Error:\\ Unsatisfied\\ constraints:.*(?:\\n.+)*.*\\(maybe\\ a\\ bugged\\ tactic\\)',
                           re_escape(error_string),
                           re.DOTALL)
    elif re.search(r'Universe [^ ]* is unbound', error_string):
        re_string = re.sub(r'Universe\\ [^ ]*\\ is\\ unbound',
                           r'Universe\\ [^ ]*\\ is\\ unbound',
                           re_escape(error_string))
    elif re.search(r'Compilation of file /tmp/[^ ]*', error_string):
        re_string = re.sub(r'Compilation\\ of\\ file\\ /tmp/[^ \.]*\\\.',
                           r'Compilation\\ of\\ file\\ /tmp/[^ \.]*\\.',
                           re_escape(error_string))
    else:
        re_string = re_escape(error_string)
    re_string = re.sub(r'tmp(?:[A-Za-z_\d]|\\_)+',
                       r'[A-Za-z_\\d\\.]+',
                       re_string)
    if r'Universe\ instance\ should\ have\ length\ ' not in re_string:
        re_string = re.sub(r'[\d]+',
                           r'[\\d]+',
                           re_string)
    ret = DEFAULT_ERROR_REG_STRING_GENERIC % unstrictify_whitespace(re_string)
    if not util.PY3: ret = ret.encode('utf-8')
    return ret

TIMEOUT = {}

def get_timeout(coqc=None):
    return TIMEOUT.get(coqc)

def set_timeout(key, value, **kwargs):
    kwargs['log']('\nThe timeout for %s has been set to: %d' % (key, value))
    TIMEOUT[key] = value

def reset_timeout():
    global TIMEOUT
    TIMEOUT = {}

def timeout_Popen_communicate(log, *args, **kwargs):
    ret = { 'value' : ('', ''), 'returncode': None }
    timeout = kwargs.get('timeout')
    del kwargs['timeout']
    input_val = kwargs.get('input')
    if input_val is not None: input_val = input_val.encode('utf-8')
    del kwargs['input']
    p = subprocess.Popen(*args, **kwargs)

    def target():
        ret['value'] = tuple(map(util.s, p.communicate(input=input_val)))
        ret['returncode'] = p.returncode

    thread = threading.Thread(target=target)
    thread.start()

    thread.join(timeout)
    if not thread.is_alive():
        return (ret['value'], ret['returncode'])

    p.terminate()
    thread.join()
    return (tuple(map((lambda s: (s if s else '') + TIMEOUT_POSTFIX), ret['value'])), ret['returncode'])


def memory_robust_timeout_Popen_communicate(log, *args, **kwargs):
    while True:
        try:
            return timeout_Popen_communicate(log, *args, **kwargs)
        except OSError as e:
            log('Warning: subprocess.Popen%s%s failed with %s\nTrying again in 10s' % (repr(tuple(args)), repr(kwargs), repr(e)), force_stdout=True, level=LOG_ALWAYS)
            time.sleep(10)

COQ_OUTPUT = {}

def sanitize_cmd(cmd):
    return re.sub(r'("/tmp/tmp)[^/"]*?((?:/[^"]*?)?(?:\.v)?")', r'\1XXXXXXXX\2', cmd)

def get_filepath_of_coq_args(coqc_prog, coqc_prog_args, **kwargs):
    coqc_help = get_coqc_help(coqc_prog, **kwargs)
    for arg_group in group_coq_args(coqc_prog_args, coqc_help):
        if arg_group[0] == '-top':
            ret = arg_group[1].split('.')
            return ret[:-1], ret[-1] + '.v'
    return None, None

prepare_cmds_for_coq_output_printed_cmd_already = set()
def prepare_cmds_for_coq_output(coqc_prog, coqc_prog_args, contents, cwd=None, timeout_val=0, **kwargs):
    def make_rmtree_onerror(file_name):
        def rmtree_onerror(function, path, exc_info):
            kwargs['log']('Non-fatal error encountered when cleaning up %s:\n' % file_name)
            etype, value, tb = exc_info
            if hasattr(traceback, 'TracebackException'):
                kwargs['log'](''.join(traceback.TracebackException(type(value), value, tb, capture_locals=True).format()))
            else:
                kwargs['log'](traceback.format_exception(type(value), value, tb))
        return rmtree_onerror

    key = (coqc_prog, tuple(coqc_prog_args), kwargs['pass_on_stdin'], contents, timeout_val, cwd)
    cmds = [coqc_prog] + list(coqc_prog_args)
    if key in COQ_OUTPUT.keys():
        file_name = COQ_OUTPUT[key][0]
        cleaner = lambda: None
    else:
        intermediate_dirs, topfilename = get_filepath_of_coq_args(coqc_prog, coqc_prog_args, **kwargs)
        if topfilename is None:
            with tempfile.NamedTemporaryFile(suffix='.v', delete=False, mode='wb') as f:
                f.write(contents.encode('utf-8'))
            file_name = f.name
            cleaner = lambda: clean_v_file(file_name)
        else:
            temp_dir_name = tempfile.mkdtemp()
            file_path = os.path.join(temp_dir_name, *intermediate_dirs)
            if not os.path.exists(file_path): os.makedirs(file_path)
            file_name = os.path.join(file_path, topfilename)
            with open(file_name, mode='wb') as f:
                f.write(contents.encode('utf-8'))
            cleaner = lambda: shutil.rmtree(temp_dir_name, onerror=make_rmtree_onerror(file_name))
            cmds.extend(['-R', temp_dir_name, '']) # make sure we bind the entire tree; use -R for compat with 8.4

    file_name_root = os.path.splitext(file_name)[0]

    pseudocmds = ''
    input_val = None
    if kwargs['is_coqtop']:
        if kwargs['pass_on_stdin']:
            input_val = contents
            cmds.extend(['-q'])
            pseudocmds = '" < "%s' % file_name
        else:
            cmds.extend(['-load-vernac-source', file_name_root, '-q'])
    else:
        cmds.extend([file_name, '-q'])
    cmd_to_print = '"%s%s"' % ('" "'.join(cmds), pseudocmds)
    kwargs['log']('\nRunning command: %s' % cmd_to_print,
                  level=(kwargs['verbose_base']
                         - (1 if sanitize_cmd(cmd_to_print) not in prepare_cmds_for_coq_output_printed_cmd_already else 0)))
    prepare_cmds_for_coq_output_printed_cmd_already.add(sanitize_cmd(cmd_to_print))
    kwargs['log']('\nContents:\n%s\n' % contents,
                  level=kwargs['verbose_base']+1)
    return key, file_name, cmds, input_val, cleaner

def reset_coq_output_cache(coqc_prog, coqc_prog_args, contents, timeout_val, cwd=None, is_coqtop=False, pass_on_stdin=False, verbose_base=1, **kwargs):
    key, file_name, cmds, input_val, cleaner = prepare_cmds_for_coq_output(coqc_prog, coqc_prog_args, contents, cwd=cwd, timeout_val=timeout_val, is_coqtop=is_coqtop, pass_on_stdin=pass_on_stdin, verbose_base=verbose_base, **kwargs)
    cleaner()

    if key in COQ_OUTPUT.keys(): del COQ_OUTPUT[key]

def get_coq_output(coqc_prog, coqc_prog_args, contents, timeout_val, cwd=None, is_coqtop=False, pass_on_stdin=False, verbose_base=1, retry_with_debug_when=(lambda output: 'is not a compiled interface for this version of OCaml' in output), **kwargs):
    """Returns the coqc output of running through the given
    contents.  Pass timeout_val = None for no timeout."""
    if timeout_val is not None and timeout_val < 0 and get_timeout(coqc_prog) is not None:
        return get_coq_output(coqc_prog, coqc_prog_args, contents, get_timeout(coqc_prog), cwd=cwd, is_coqtop=is_coqtop, pass_on_stdin=pass_on_stdin, verbose_base=verbose_base, retry_with_debug_when=retry_with_debug_when, **kwargs)

    key, file_name, cmds, input_val, cleaner = prepare_cmds_for_coq_output(coqc_prog, coqc_prog_args, contents, cwd=cwd, timeout_val=timeout_val, is_coqtop=is_coqtop, pass_on_stdin=pass_on_stdin, verbose_base=verbose_base, **kwargs)

    if key in COQ_OUTPUT.keys():
        cleaner()
        return COQ_OUTPUT[key][1]

    start = time.time()
    ((stdout, stderr), returncode) = memory_robust_timeout_Popen_communicate(kwargs['log'], cmds, stderr=subprocess.STDOUT, stdout=subprocess.PIPE, stdin=subprocess.PIPE, timeout=(timeout_val if timeout_val is not None and timeout_val > 0 else None), input=input_val, cwd=cwd)
    finish = time.time()
    runtime = finish - start
    kwargs['log']('\nretcode: %d\nstdout:\n%s\n\nstderr:\n%s\nruntime:\n%f\n\n' % (returncode, util.s(stdout), util.s(stderr), runtime),
                  level=verbose_base + 1)
    if get_timeout(coqc_prog) is None and timeout_val is not None:
        set_timeout(coqc_prog, 3 * max((1, int(math.ceil(finish - start)))), **kwargs)
    cleaner()
    COQ_OUTPUT[key] = (file_name, (clean_output(util.s(stdout)), tuple(cmds), returncode, runtime))
    kwargs['log']('Storing result: COQ_OUTPUT[%s]:\n%s' % (repr(key), repr(COQ_OUTPUT[key])),
                  level=verbose_base + 2)
    if retry_with_debug_when(COQ_OUTPUT[key][1][0]):
        debug_args = get_coq_debug_native_compiler_args(coqc_prog)
        kwargs['log']('Retrying with %s...' % ' '.join(debug_args),
                      level=verbose_base - 1)
        return get_coq_output(coqc_prog, list(debug_args) + list(coqc_prog_args), contents, timeout_val, cwd=cwd, is_coqtop=is_coqtop, pass_on_stdin=pass_on_stdin, verbose_base=verbose_base, retry_with_debug_when=(lambda output: False), **kwargs)
    return COQ_OUTPUT[key][1]



from __future__ import with_statement
import subprocess, tempfile, re, os
from .file_util import clean_v_file
from .memoize import memoize
from . import util

__all__ = ["get_coqc_version", "get_coqtop_version", "get_coqc_help", "get_coqc_coqlib", "get_coq_accepts_top", "get_coq_accepts_time", "get_coq_accepts_emacs", "get_coq_accepts_o", "get_coq_accepts_compile", "get_coq_native_compiler_ondemand_fragment", "group_coq_args_split_recognized", "group_coq_args", "coq_makefile_supports_arg", "DEFAULT_COQTOP"]

# {Windows,Python,coqtop} is terrible; we fail to write to (or read
# from?) coqtop.  But we can wrap it in a batch script, and it works
# fine.
DEFAULT_COQTOP = 'coqtop' if os.name != 'nt' else util.resource_path('coqtop.bat')

@memoize
def get_coqc_version_helper(coqc):
    p = subprocess.Popen([coqc, "-q", "-v"], stderr=subprocess.STDOUT, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    return util.normalize_newlines(util.s(stdout).replace('The Coq Proof Assistant, version ', '')).replace('\n', ' ').strip()

def get_coqc_version(coqc_prog, **kwargs):
    kwargs['log']('Running command: "%s"' % '" "'.join([coqc_prog, "-q", "-v"]), level=2)
    return get_coqc_version_helper(coqc_prog)

@memoize
def get_coqc_help_helper(coqc):
    p = subprocess.Popen([coqc, "-q", "--help"], stderr=subprocess.STDOUT, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    return util.s(stdout).strip()

def get_coqc_help(coqc_prog, **kwargs):
    kwargs['log']('Running command: "%s"' % '" "'.join([coqc_prog, "-q", "--help"]), level=2)
    return get_coqc_help_helper(coqc_prog)

@memoize
def get_coqtop_version_helper(coqtop):
    p = subprocess.Popen([coqtop, "-q"], stderr=subprocess.PIPE, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    return util.normalize_newlines(util.s(stdout).replace('Welcome to Coq ', '').replace('Skipping rcfile loading.', '')).replace('\n', ' ').strip()

def get_coqtop_version(coqtop_prog, **kwargs):
    kwargs['log']('Running command: "%s"' % '" "'.join([coqtop_prog, "-q", "-v"]), level=2)
    return get_coqtop_version_helper(coqtop_prog)

@memoize
def get_coq_accepts_top(coqc):
    temp_file = tempfile.NamedTemporaryFile(suffix='.v', dir='.', delete=True)
    temp_file_name = temp_file.name
    p = subprocess.Popen([coqc, "-q", "-top", "Top", temp_file_name], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    temp_file.close()
    clean_v_file(temp_file_name)
    return '-top: no such file or directory' not in util.s(stdout)

@memoize
def get_coq_accepts_compile(coqtop):
    temp_file = tempfile.NamedTemporaryFile(suffix='.v', dir='.', delete=True)
    temp_file_name = temp_file.name
    p = subprocess.Popen([coqtop, "-q", "-compile", temp_file_name], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    rc = p.returncode
    temp_file.close()
    clean_v_file(temp_file_name)
    return rc == 0

def get_coq_accepts_option(coqc_prog, option, **kwargs):
    help_text = get_coqc_help(coqc_prog, **kwargs)
    return any((option + sep) in help_text for sep in '\t ')

def get_coq_accepts_o(coqc_prog, **kwargs):
    return get_coq_accepts_option(coqc_prog, '-o', **kwargs)

def get_coq_accepts_time(coqc_prog, **kwargs):
    return get_coq_accepts_option(coqc_prog, '-time', **kwargs)

def get_coq_accepts_emacs(coqc_prog, **kwargs):
    return get_coq_accepts_option(coqc_prog, '-emacs', **kwargs)

def get_coq_accepts_w(coqc_prog, **kwargs):
    return get_coq_accepts_option(coqc_prog, '-w', **kwargs)

@memoize
def get_coqc_native_compiler_ondemand_errors(coqc):
    temp_file = tempfile.NamedTemporaryFile(suffix='.v', dir='.', delete=True)
    temp_file_name = temp_file.name
    p = subprocess.Popen([coqc, "-q", "-native-compiler", "ondemand", temp_file_name], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    temp_file.close()
    clean_v_file(temp_file_name)
    return any(v in util.s(stdout) for v in ('The native-compiler option is deprecated',
                                             'Native compiler is disabled',
                                             'native-compiler-disabled',
                                             'deprecated-native-compiler-option'))

def get_coq_native_compiler_ondemand_fragment(coqc_prog, **kwargs):
    help_lines = get_coqc_help(coqc_prog, **kwargs).split('\n')
    if any('ondemand' in line for line in help_lines if line.strip().startswith('-native-compiler')):
        if get_coq_accepts_w(coqc_prog, **kwargs):
            return ('-w', '-deprecated-native-compiler-option,-native-compiler-disabled', '-native-compiler', 'ondemand')
        elif not get_coqc_native_compiler_ondemand_errors(coqc_prog):
            return ('-native-compiler', 'ondemand')
    return tuple()

def get_coq_accepts_time(coqc_prog, **kwargs):
    return '-time' in get_coqc_help(coqc_prog, **kwargs)

HELP_REG = re.compile(r'^  ([^\n]*?)(?:\t|  )', re.MULTILINE)
HELP_MAKEFILE_REG = re.compile(r'^\[(-[^\n\]]*)\]', re.MULTILINE)

def adjust_help_string(coqc_help):
    # we need to insert some spaces to make parsing easier
    for one_arg_option in ('-diffs', '-native-compiler'):
        coqc_help = re.sub(r'(\n\s*%s\s+[^\s]*)' % one_arg_option, r'\1 ', coqc_help, re.MULTILINE)
    return coqc_help

def all_help_tags(coqc_help, is_coq_makefile=False):
    coqc_help = adjust_help_string(coqc_help)
    if is_coq_makefile:
        return HELP_MAKEFILE_REG.findall(coqc_help)
    else:
        return HELP_REG.findall(coqc_help)

def get_single_help_tags(coqc_help, **kwargs):
    return tuple(t
                 for i in all_help_tags(coqc_help, **kwargs) if ' ' not in i.replace(', ', '')
                 for t in i.split(', ') if len(t) > 0)

def get_multiple_help_tags(coqc_help, **kwargs):
    return dict((t.split(' ')[0], len(t.split(' ')))
                for i in all_help_tags(coqc_help, **kwargs) if ' ' in i.replace(', ', '')
                for t in re.sub(r'"[^"]*"', '""', i).split(', '))

def coq_makefile_supports_arg(coq_makefile_help):
    tags = get_multiple_help_tags(coq_makefile_help, is_coq_makefile=True)
    return any(tag == '-arg' for tag in tags.keys())

def group_coq_args_split_recognized(args, coqc_help, topname=None, is_coq_makefile=False):
    args = list(args)
    bindings = []
    unrecognized_bindings = []
    single_tags = get_single_help_tags(coqc_help, is_coq_makefile=is_coq_makefile)
    multiple_tags = get_multiple_help_tags(coqc_help, is_coq_makefile=is_coq_makefile)
    while len(args) > 0:
        if args[0] in multiple_tags.keys() and len(args) >= multiple_tags[args[0]]:
            cur_binding, args = tuple(args[:multiple_tags[args[0]]]), args[multiple_tags[args[0]]:]
            if cur_binding not in bindings:
                bindings.append(cur_binding)
        else:
            cur = tuple([args.pop(0)])
            if cur[0] not in single_tags:
                unrecognized_bindings.append(cur[0])
            elif cur not in bindings:
                bindings.append(cur)
    if topname is not None and '-top' not in [i[0] for i in bindings] and '-top' in multiple_tags.keys():
        bindings.append(('-top', topname))
    return (bindings, unrecognized_bindings)

def group_coq_args(args, coqc_help, topname=None, is_coq_makefile=False):
    bindings, unrecognized_bindings = group_coq_args_split_recognized(args, coqc_help, topname=topname, is_coq_makefile=is_coq_makefile)
    return bindings + [tuple([v]) for v in unrecognized_bindings]

@memoize
def get_coqc_config_helper(coqc, coq_args):
    p = subprocess.Popen([coqc, "-q", "-config"] + list(coq_args), stderr=subprocess.STDOUT, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    return util.normalize_newlines(util.s(stdout)).strip()

def get_coqc_config(coqc_prog, coq_args=tuple(), **kwargs):
    grouped_args = group_coq_args(coq_args, get_coqc_help(coqc_prog, **kwargs))
    coq_args = [arg for args in grouped_args if args[0] in ('--coqlib', '-coqlib') for arg in args]
    kwargs['log']('Running command: "%s"' % '" "'.join([coqc_prog, "-q", "-config"] + coq_args), level=2)
    return get_coqc_config_helper(coqc_prog, tuple(coq_args))

def get_coqc_coqlib(coqc_prog, **kwargs):
    return [line[len('COQLIB='):] for line in get_coqc_config(coqc_prog, **kwargs).split('\n') if line.startswith('COQLIB=')][0]



from setuptools import setup, find_packages

setup(
    name='coq-tools',
    version='0.0.1',
    packages=find_packages(),
    install_requires=[
        # Add any dependencies here
    ],
    scripts=[
        # Add non-python executable scripts to install here
    ],
    package_data={
        'coq_tools': [
            'coqtop-as-coqc.sh',
            'coqtop.bat',
        ],
    },
    entry_points={
        'console_scripts': [
            'coq-bug-minimizer=coq_tools.find_bug:main',
            'coq-require-minimizer=coq_tools.minimize_requires:main',
            'coq-import-inliner=coq_tools.inline_imports:main',
        ],
    }
)
